/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "GServClient.h"
#include "SharedMem.h"
#include "NewClipService.h"
#include "Globals.h"
#include "DbgPrintf.h"
#include "CriticalSection.h"
#include "ParamsUtil.h"
#include "HostCalls.h"
#include "DlgWaitProgress.h"

#include "LangTools.h"
#include "VBoxUtil.h"
#include "SecureFuncs.h"

#include <strsafe.h>

#include <vector>
#include <set>
#include <string>
#include "Settings.h"


using namespace std;

#ifndef MSGFLT_ADD

#define MSGFLT_ADD 1

#endif

#define MAX_WAIT_FOR_INPUT_ITEMS (1)

typedef struct _WAIT_FOR_ANSWER_ITEM
{
    UINT    ErrorCode;

    HANDLE  hEvent;
    UINT    ClientId;
    UINT    PacketId;

    VBOXHGCMCALLHANDLE callHandle;
    VBOXHGCMSVCPARM *paParms;
    PVBOXHGCMSVCHELPERS pHelpers;

}WAIT_FOR_ANSWER_ITEM, *PWAIT_FOR_ANSWER_ITEM;

typedef struct _WAIT_FOR_INPUT_ITEM
{
    PVBOXHGCMSVCHELPERS pHelpers;
    VBOXHGCMCALLHANDLE callHandle;
    VBOXHGCMSVCPARM *paParms;
    LONG PacketId;
    uint32_t u32ClientID;

} WAIT_FOR_INPUT_ITEM;

DWORD
WaitForAnswerFromGuestServer( HANDLE hEvent );

BOOL
WINAPI
Vista_ChangeWindowMessageFilter(
    UINT message,
    DWORD dwFlag
    );


VOID
OnClientDisconnect( uint32_t u32ClientID );


VOID
Host_PasteFilesToFolder(
    LPWSTR lpFolderPath
    );

VOID
CreateContextMenuCmdWorkerThread();

DWORD
WINAPI
ContextMenuCmdWorkerThread(
    LPVOID lpParameter
    );

LRESULT
CALLBACK
ContextMenuCmdWorkerWndProc(
    HWND hwnd, 
    UINT msg, 
    WPARAM wParam, 
    LPARAM lParam
    );


extern wstring g_MachineName;
HWND g_CreateContextMenuCmdWorkerWnd = NULL;
CSharedMem g_SharedMemCmdWindoInfo;
SHARED_MEM_SHELLMENU_COMMAND_WINDOW *g_pShellMenuCommandWindow = NULL;
volatile LONG g_PacketId  = 1;
UINT g_GuestClipboardOwnerId = 0;

vector<WAIT_FOR_INPUT_ITEM> g_WaitForInputItems;
CCriticalSection g_WaitForInputItemsLock;

vector<PWAIT_FOR_ANSWER_ITEM> g_WaitForAnswerItems;
CCriticalSection g_WaitForAnswerItemsLock;

set<uint32_t> g_ClientIds;


BOOL
GSERVCLIENT_Handle_svcCall(
    PVBOXHGCMSVCHELPERS pHelpers,
    void *ptrUnk,
    VBOXHGCMCALLHANDLE callHandle,
    uint32_t u32ClientID,
    void *pvClient,
    uint32_t u32Function,
    uint32_t cParms,
    VBOXHGCMSVCPARM paParms[]
    )
{
    //pHelpers->pfnCallComplete ( callHandle, VINF_SUCCESS );
	
	DbgPrintf( "GSERVCLIENT_Handle_svcCall pHelpers=%p u32ClientID=%u u32Function=0x%x\n", pHelpers, u32ClientID, u32Function );

    switch ( u32Function )
    {
        case VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES:
        {
            //TODO: Save in shared memory clipboard window handle & pid

            DbgPrintf( "VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES processing\n" );

            if ( cParms != VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES_CPARAMS )
            {
                DbgPrintf( "VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES: invalid parameters count\n" );
                pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
                return TRUE;
            }

            UINT OSVersion = 0;
            UINT Is64BitOS = 0;

            if ( !VBoxUINT32ParamGet( &paParms[0], &OSVersion ) ||
                 !VBoxUINT32ParamGet( &paParms[1], &Is64BitOS )
                )
            {
                DbgPrintf( "VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES: invalid parameters type\n" );
                pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
                return TRUE;
            }


            if ( g_pShellMenuCommandWindow )
            {
                g_pShellMenuCommandWindow->CommandWnd = (ULONG64)g_CreateContextMenuCmdWorkerWnd;
                g_pShellMenuCommandWindow->uProcessId = GetCurrentProcessId();
                g_pShellMenuCommandWindow->GuestOSVersion = OSVersion;
                g_pShellMenuCommandWindow->GuestOSIs64bit = Is64BitOS;
				
				StringCchCopyW( 
					g_pShellMenuCommandWindow->wszMachineName,
					ARRAYSIZE(g_pShellMenuCommandWindow->wszMachineName),
					g_MachineName.c_str()
					);

                DbgPrintf( "GUESTOS Ver: 0x%x 64bit: %u\n", OSVersion, Is64BitOS );
            }

            g_GuestClipboardOwnerId = u32ClientID;

            pHelpers->pfnCallComplete ( callHandle, VINF_SUCCESS );
            return TRUE;
            break;
        }
        
        case VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER:
        {
            if ( cParms != VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER_CPARAMS )
            {
				DbgPrintf( "cParms != VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER_CPARAMS\n" );
                pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
                return TRUE;
            }

            if ( g_WaitForInputItems.size() >= MAX_WAIT_FOR_INPUT_ITEMS )
            {
                DbgPrintf( "Error: g_WaitForInputItems.size(%u) >= MAX_WAIT_FOR_INPUT_ITEMS\n", (ULONG)g_WaitForInputItems.size() );
                pHelpers->pfnCallComplete ( callHandle, VERR_ACCESS_DENIED );
                return TRUE;
            }

            WAIT_FOR_INPUT_ITEM NewItem;

            NewItem.PacketId = InterlockedIncrement( &g_PacketId );
            NewItem.paParms = &paParms[0];
            NewItem.callHandle = callHandle;
            NewItem.u32ClientID = u32ClientID;
            NewItem.pHelpers = pHelpers;
			
			DbgPrintf( "NewWaitInp: PktId=%u Client=%u callHandle=0x%x\n", g_PacketId, u32ClientID, (ULONG)(ULONG_PTR)callHandle );

            g_WaitForInputItemsLock.Lock();
            g_WaitForInputItems.push_back( NewItem );
            g_WaitForInputItemsLock.Unlock();

            return TRUE;
            break;
        }

        case VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER:
        {
            if ( cParms != VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER_CPARAMS )
            {
				DbgPrintf( "cParms != VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER_CPARAMS\n" );
                pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
                return TRUE;
            }

            //ErrorCode PacketId Buffer

            UINT PacketId = 0;
            UINT ErrorCode = 0;


            if ( paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
                 paParms[1].type != VBOX_HGCM_SVC_PARM_32BIT ||
                 paParms[2].type != VBOX_HGCM_SVC_PARM_PTR ||
                 !VBoxUINT32ParamGet( &paParms[0], &ErrorCode ) ||
                 !VBoxUINT32ParamGet( &paParms[1], &PacketId )
                )
            {
				DbgPrintf( "VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER_CPARAMS: Bad params type\n" );
                pHelpers->pfnCallComplete ( callHandle, VERR_INVALID_PARAMETER );
                return TRUE;
            }

            BOOL bFound = FALSE;
            g_WaitForAnswerItemsLock.Lock();

            for (
                vector<PWAIT_FOR_ANSWER_ITEM>::iterator iter = g_WaitForAnswerItems.begin();
                iter != g_WaitForAnswerItems.end();
                iter++
                )
            {
                if ( (*iter)->ClientId == u32ClientID &&
                     (*iter)->PacketId == PacketId
                   )
                {
                    bFound = TRUE;

                    (*iter)->ErrorCode = ErrorCode;
                    (*iter)->paParms = &paParms[0];
                    (*iter)->callHandle = callHandle;
                    (*iter)->pHelpers = pHelpers;
					
					DbgPrintf( "VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER set event: %p pkt=%u\n", (*iter)->hEvent, PacketId );

                    SetEvent( (*iter)->hEvent );
                    break;
                }

            } //for

            g_WaitForAnswerItemsLock.Unlock();

            if ( !bFound )
            {
                DbgPrintf( "ERROR: Unknown output packet(id=%u)\n", PacketId );
                pHelpers->pfnCallComplete ( callHandle, VERR_GENERAL_FAILURE );
            }
            

            return TRUE;
            break;
        }

    }

    return FALSE;
}

VOID
CreateContextMenuCmdWorkerThread()
{
    DWORD dwTid = 0;
    static BOOL bWasInit = FALSE;

    if ( bWasInit )
        return;

    HANDLE hThread = CreateThread(
        NULL,
        0,
        (LPTHREAD_START_ROUTINE)ContextMenuCmdWorkerThread,
        NULL,
        0,
        &dwTid
        );

    DWORD dwGle = GetLastError();

    if ( hThread == NULL )
    {
        DbgPrintf( "Cannot create thread: %u\n", dwGle );
    }
    else
    {
        CloseHandle( hThread );
    }

    g_pShellMenuCommandWindow = (SHARED_MEM_SHELLMENU_COMMAND_WINDOW *)g_SharedMemCmdWindoInfo.Create(
        SHARED_MEM_NAMEA,
        sizeof(SHARED_MEM_SHELLMENU_COMMAND_WINDOW),
        NULL
        );

    bWasInit = TRUE;

}

DWORD
WINAPI
ContextMenuCmdWorkerThread(
    LPVOID lpParameter
    )
{
    WNDCLASSEX wcex;
    HINSTANCE hInstance = GetModuleHandle(NULL);

    RtlZeroMemory( &wcex, sizeof(wcex) );

    wcex.cbSize        = sizeof(WNDCLASSEX);
    wcex.style         = 0;
    wcex.lpfnWndProc   = ContextMenuCmdWorkerWndProc;
    wcex.hInstance     = hInstance;
    wcex.hIcon         = NULL;
    wcex.hCursor       = NULL;
    wcex.hbrBackground = NULL;
    wcex.lpszMenuName  = NULL;
    wcex.lpszClassName = "ContextMenuCmdWorkerClass";

    if( !RegisterClassEx( &wcex ) )
    {
        DbgPrintf( "RegisterClassEx failed: %u\n", GetLastError() );
        return 0;
    }

    HWND hWnd = CreateWindowEx(
        0,
        wcex.lpszClassName,
        "",
        WS_OVERLAPPEDWINDOW,
        0, 0, 0, 0,
        NULL,
        NULL, 
        hInstance,
        NULL
        );

    if ( NULL == hWnd )
    {
        DbgPrintf( "CreateWindowEx failed: %u\n", GetLastError() );
        return 0;
    }

    g_CreateContextMenuCmdWorkerWnd = hWnd;

    //FIXME?
    Vista_ChangeWindowMessageFilter(
        WM_COPYDATA,
        MSGFLT_ADD
        );

    ShowWindow( hWnd, SW_HIDE );

    MSG Msg;

    while( GetMessage( &Msg, NULL, 0, 0) )
    {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
    }

    return 0;
}


LRESULT
CALLBACK
ContextMenuCmdWorkerWndProc(
    HWND hwnd, 
    UINT msg, 
    WPARAM wParam, 
    LPARAM lParam
    )
{
    switch( msg )
    {
        case WM_COPYDATA:
        {
            PCOPYDATASTRUCT pCopyData = (PCOPYDATASTRUCT)lParam;

            if ( pCopyData->dwData == SHARED_MEM_SHELLMENU_COPY_DATA_PARAM && 
                 pCopyData->cbData == sizeof(SHARED_MEM_SHELLMENU_CMD_PARAMS_PASTE) 
                )
            {
                SHARED_MEM_SHELLMENU_CMD_PARAMS_PASTE CapturedShellMenuInfo;

                RtlCopyMemory( &CapturedShellMenuInfo, pCopyData->lpData, sizeof(CapturedShellMenuInfo) );
                CapturedShellMenuInfo.wszFolder[ MAX_PATH - 1 ] = 0; //ensure null terminator

                Host_PasteFilesToFolder( CapturedShellMenuInfo.wszFolder );

            }

            break;
        }

        case WM_DESTROY:
        {
            PostQuitMessage(0);
            break;
        }

        case WM_CLOSE:
        {
            DestroyWindow( hwnd );
            break;
        }

        default:
        {
            return DefWindowProc(hwnd, msg, wParam, lParam);
        }
    }
    return 0;
}

VOID
OnClientDisconnect( uint32_t u32ClientID )
{
   if ( u32ClientID == g_GuestClipboardOwnerId )
    {
        g_GuestClipboardOwnerId = 0;

        if ( g_pShellMenuCommandWindow && g_pShellMenuCommandWindow->uProcessId == GetCurrentProcessId() )
        {
            //g_pShellMenuCommandWindow->uProcessId = 0;
            //g_pShellMenuCommandWindow->CommandWnd = 0;
            RtlZeroMemory( g_pShellMenuCommandWindow, sizeof(SHARED_MEM_SHELLMENU_COMMAND_WINDOW) );
        }
    }

    g_WaitForInputItemsLock.Lock();

    for (
        vector<WAIT_FOR_INPUT_ITEM>::iterator iter = g_WaitForInputItems.begin();
        iter != g_WaitForInputItems.end();
        )
    {
        if ( iter->u32ClientID == u32ClientID )
        {
            iter->pHelpers->pfnCallComplete( iter->callHandle, VERR_INTERRUPTED );

            iter = g_WaitForInputItems.erase( iter );
            continue;
        }

        iter++;
    }

    g_WaitForInputItemsLock.Unlock();

    g_WaitForAnswerItemsLock.Lock();

    //remove

    for (
        vector<PWAIT_FOR_ANSWER_ITEM>::iterator iter = g_WaitForAnswerItems.begin();
        iter != g_WaitForAnswerItems.end();
        iter++)
    {
        if ( (*iter)->ClientId == u32ClientID )
        {
            (*iter)->ErrorCode = ERROR_OPERATION_ABORTED;

            if ( (*iter)->hEvent != NULL ) //ADDED 19.06 - Free resource
            {
                SetEvent( (*iter)->hEvent );
                CloseHandle( (*iter)->hEvent ); //..
                (*iter)->hEvent = NULL;
            }


        }
    }

    g_WaitForAnswerItemsLock.Unlock();

}

VOID
GSERVCLIENT_Handle_svcDisconnect(
    void *pUnk1, 
    uint32_t u32ClientID, 
    void *pvClient
    )
{
    OnClientDisconnect( u32ClientID );
	
	g_ClientIds.erase( u32ClientID );
	
}

VOID
GSERVCLIENT_Handle_svcSaveState(
    void *pvService, 
    uint32_t u32ClientID, 
    void *pvClient, 
    PSSMHANDLE pSSM
    )
{
    //OnClientDisconnect( u32ClientID );
    //saving state does not mean exiting of client
	
	DbgPrintf( "svcSaveState called for %u\n", u32ClientID );
	
	/*g_WaitForInputItemsLock.Lock();

    for (
        vector<WAIT_FOR_INPUT_ITEM>::iterator iter = g_WaitForInputItems.begin();
        iter != g_WaitForInputItems.end();
        )
    {
        if ( iter->u32ClientID == u32ClientID )
        {
			//VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER
			
			DbgPrintf( "svcSaveState(%u): removing pending requests\n", u32ClientID );
						
			VBoxUINT32ParamSet( &iter->paParms[0], ERROR_GEN_FAILURE );//FIXME
            iter->pHelpers->pfnCallComplete( iter->callHandle,  VINF_SUCCESS );	//VERR_INTERRUPTED
			
			//DbgPrintf( "FIXME: pfnDisconnectClient( pHelpers=%p )\n", iter->pHelpers );
			//iter->pHelpers->pfnDisconnectClient( iter->pHelpers->pvInstance, u32ClientID ); //FIXME			

            iter = g_WaitForInputItems.erase( iter );	
			
        }
		else
		{
			iter++;
		}
    }

    g_WaitForInputItemsLock.Unlock();*/
}



VOID
GS_FreeMem( PVOID pPtr )
{
    free( pPtr );
}

UINT
GS_GetClipboardOwnerClientId()
{
    return g_GuestClipboardOwnerId;
}



UINT
CallGuestServer(
     IN  ULONG ClientId,
     IN  PVOID pInputBuffer,
     IN  ULONG InputBufferSize,
     OUT PVOID *ppOutputBuffer,
     OUT  ULONG *pOutputBufferSize
     )
{
    WAIT_FOR_INPUT_ITEM FoundItem;
    BOOL bFound = FALSE;
    int TryCount = 15;

    RtlZeroMemory( &FoundItem, sizeof(FoundItem) );

    do
    {
        g_WaitForInputItemsLock.Lock();

        for (
            vector<WAIT_FOR_INPUT_ITEM>::iterator iter = g_WaitForInputItems.begin();
            iter != g_WaitForInputItems.end();
            )
        {
            if ( iter->u32ClientID == (UINT)ClientId )
            {
                FoundItem = *iter;
                iter = g_WaitForInputItems.erase( iter );
                bFound = TRUE;
                break;
            }

            iter++;
        }
        
        g_WaitForInputItemsLock.Unlock();

        if ( !bFound )
        {
            Sleep(100);
        }

        TryCount--;

    } while ( !bFound && TryCount > 0 );

    if ( !bFound )
    {
        DbgPrintf( "CallGuestServer: Cannot find client connection(%u) Clients=%u\n", ClientId, (ULONG)g_WaitForInputItems.size() );
        return ERROR_INVALID_HANDLE;
    }

    //HGCMFunctionParameter ErrorCode; //OUT UINT
    //HGCMFunctionParameter Buffer; //IN OUT PTR
    //HGCMFunctionParameter OutputBufferSize; //OUT
    //HGCMFunctionParameter PacketId; //OUT UINT

    if ( FoundItem.paParms[0].type != VBOX_HGCM_SVC_PARM_32BIT ||
         FoundItem.paParms[1].type != VBOX_HGCM_SVC_PARM_PTR ||
         FoundItem.paParms[2].type != VBOX_HGCM_SVC_PARM_32BIT ||
         FoundItem.paParms[3].type != VBOX_HGCM_SVC_PARM_32BIT
       )
    {
        DbgPrintf( "CallGuestServer: Invalid param type\n" );
        FoundItem.pHelpers->pfnCallComplete( FoundItem.callHandle, VERR_INVALID_PARAMETER );
        return ERROR_INVALID_PARAMETER;
    }

    PVOID pPtr = NULL;
    uint32_t PtrSize = 0;

    if ( !VBoxPointerParamGet( &FoundItem.paParms[1], &pPtr, &PtrSize ) ||
          PtrSize < InputBufferSize
        )
    {
        DbgPrintf( "CallGuestServer: Invalid input buffer param\n" );
        FoundItem.pHelpers->pfnCallComplete( FoundItem.callHandle, VERR_INVALID_PARAMETER );
        return ERROR_INVALID_PARAMETER;
    }

    RtlCopyMemory( pPtr, pInputBuffer, InputBufferSize );

    VBoxUINT32ParamSet( &FoundItem.paParms[2], InputBufferSize );
    VBoxUINT32ParamSet( &FoundItem.paParms[3], (UINT)FoundItem.PacketId );


    VBoxUINT32ParamSet( &FoundItem.paParms[0], ERROR_SUCCESS );

    //wait for answer 

    WAIT_FOR_ANSWER_ITEM NewItem;

    RtlZeroMemory( &NewItem, sizeof(NewItem) );

    NewItem.ClientId = ClientId;
    NewItem.PacketId = FoundItem.PacketId;
    NewItem.ErrorCode = ERROR_GEN_FAILURE;
    NewItem.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL );

    if ( NULL == NewItem.hEvent )
    {
        DbgPrintf( "Error: Cannot create event\n" );
        FoundItem.pHelpers->pfnCallComplete( FoundItem.callHandle, VERR_GENERAL_FAILURE );
        return ERROR_GEN_FAILURE;
    }

    g_WaitForAnswerItemsLock.Lock();
    g_WaitForAnswerItems.push_back( &NewItem );
    g_WaitForAnswerItemsLock.Unlock();

    FoundItem.pHelpers->pfnCallComplete( FoundItem.callHandle, VINF_SUCCESS );

    DWORD dwWaitResult =  WaitForAnswerFromGuestServer(
        NewItem.hEvent
        );

    if ( dwWaitResult == WAIT_TIMEOUT )
    {
        DbgPrintf( "CallGuestServer: Error: WaitForSingleObject returns WAIT_TIMEOUT(CID=%u,PKT=%u)\n", ClientId, FoundItem.PacketId );
    }
    else if ( dwWaitResult != WAIT_OBJECT_0 )
    {
        DbgPrintf( "CallGuestServer: Error: WaitForSingleObject returns %u\n", dwWaitResult );
    }

    g_WaitForAnswerItemsLock.Lock();
    //remove

    for (
        vector<PWAIT_FOR_ANSWER_ITEM>::iterator iter = g_WaitForAnswerItems.begin();
        iter != g_WaitForAnswerItems.end();
        iter++)
    {
        if ( *iter == &NewItem )
        {
            g_WaitForAnswerItems.erase( iter );
            break;
        }
    }

    if ( NewItem.hEvent ) //FIXED
    {
        CloseHandle( NewItem.hEvent );
        NewItem.hEvent = NULL;
    }

    g_WaitForAnswerItemsLock.Unlock();


    if ( NewItem.ErrorCode != ERROR_SUCCESS )
    {
		DbgPrintf( "CallGuestServer: Error: NewItem.ErrorCode: %u\n", NewItem.ErrorCode );
		
        if ( NewItem.paParms && NewItem.pHelpers && NewItem.callHandle )
        {
            NewItem.pHelpers->pfnCallComplete( NewItem.callHandle, VINF_SUCCESS );
        }

        return NewItem.ErrorCode;
    }

    PtrSize = 0;

    if ( !VBoxPointerParamGet( &NewItem.paParms[2], &pPtr, &PtrSize ) )
    {
		DbgPrintf( "CallGuestServer: VBoxPointerParamGet(NewItem.paParms[2]) failed\n" );
		
        if ( NewItem.paParms && NewItem.pHelpers && NewItem.callHandle )
        {
            NewItem.pHelpers->pfnCallComplete( NewItem.callHandle, VERR_INVALID_PARAMETER );
        }

        return ERROR_INVALID_PARAMETER;
    }

    PVOID pNewPtr = malloc(PtrSize+1);

    if ( NULL == pNewPtr )
    {
        if ( NewItem.paParms && NewItem.pHelpers && NewItem.callHandle )
        {
            NewItem.pHelpers->pfnCallComplete( NewItem.callHandle, VERR_INVALID_PARAMETER );
        }

        return ERROR_OUTOFMEMORY;
    }

    RtlCopyMemory( pNewPtr, pPtr, PtrSize );

    // free buffers now

    if ( NewItem.paParms && NewItem.pHelpers && NewItem.callHandle )
    {
        NewItem.pHelpers->pfnCallComplete( NewItem.callHandle, VINF_SUCCESS );
    }
	else
	{
		DbgPrintf( "CallGuestServer cannot pfnCallComplete(%p,%p,%p)\n", NewItem.paParms, NewItem.pHelpers, NewItem.callHandle );
	}


    *ppOutputBuffer = pNewPtr;
    *pOutputBufferSize = PtrSize;

    return ERROR_SUCCESS;
}


VOID
Host_PasteFilesToFolder(
    LPWSTR lpFolderPath
    )
{
    DbgPrintf( "[VBOX HookDll]: Paste files to folder: %ws\n", lpFolderPath );

    if ( 0 == g_GuestClipboardOwnerId )
    {
        DbgPrintf( "Guestlipboard owner is not definned\n" );
        return;
    }
	
	UINT ClipBoardMode = CLIPBOARD_MODE_BIDIRECTIONAL;
	
	if ( VBoxUtil_GetClipboardMode( &ClipBoardMode ) &&
		!( ClipBoardMode == CLIPBOARD_MODE_BIDIRECTIONAL || ClipBoardMode == CLIPBOARD_MODE_GUEST_TO_HOST )
		)
	{
		LangErrorMessageBox( IDS_CLIP_COPY_IS_NOT_ALLOWED );
		return;
	}

    CHostFromGuestCopierHelper CopyHelper;

    ULONG uRet = ProceeedCopyingFiles(
        &CopyHelper,
        g_GuestClipboardOwnerId,
        0, //DropHandle
        lpFolderPath
        );
		
	if ( uRet > 0 && Settings_BeepAfterCopyingFiles() )
	{
		MessageBeep( MB_OK );
	}
		
		
}

BOOL
WINAPI
Vista_ChangeWindowMessageFilter(
    UINT message,
     DWORD dwFlag
    )
{
    typedef
    BOOL
    (WINAPI *TChangeWindowMessageFilter)(
        UINT message,
        DWORD dwFlag
        );

    HINSTANCE hUser32 = GetModuleHandle( "user32.dll" );

    if ( NULL == hUser32 )
        return FALSE;

    TChangeWindowMessageFilter pfnChangeWindowMessageFilter = (TChangeWindowMessageFilter)GetProcAddress(
        hUser32,
        "ChangeWindowMessageFilter"
        );

    if ( NULL == pfnChangeWindowMessageFilter )
        return FALSE;

    return pfnChangeWindowMessageFilter( message, dwFlag );
}

UINT
GetWaitForInputFirstClientId()
{
    UINT ClientId = 0;

    g_WaitForInputItemsLock.Lock();

    vector<WAIT_FOR_INPUT_ITEM>::iterator iter = g_WaitForInputItems.begin();

    if ( iter != g_WaitForInputItems.end() )
    {
        ClientId = iter->u32ClientID;

        if ( g_WaitForInputItems.size() != 1 )
        {
            DbgPrintf( "WARNING: GetWaitForInputFirstClientId: several clients are in queue\n" );
        }

    }

    g_WaitForInputItemsLock.Unlock();
    return ClientId;
}

VOID
GSERVCLIENT_Handle_svcLoadState(
    void *pvService, 
    uint32_t u32ClientID, 
    void *pvClient, 
    PSSMHANDLE pSSM
    )
{
	DbgPrintf( "svcLoadState is called for %u\n",  u32ClientID );
}


VOID
GSERVCLIENT_Handle_svcConnect(
	void *pvService,
	uint32_t u32ClientID,
	void *pvClient
	)
{
	g_ClientIds.insert( u32ClientID );
	
	DbgPrintf( "svcConnect is called for %u\n",  u32ClientID );
}

DWORD
WaitForAnswerFromGuestServer( HANDLE hEvent )
{
	//WAIT_TIMEOUT WAIT_OBJECT_0 WAIT_FAILED
	
	DWORD dwTimeOut = GetTickCount() + TIMEOUT_GS_WAIT_FOR_GSERVER_RESPONSE;
	
	DWORD dwRet = WaitForSingleObject(
		hEvent,
		TIMEOUT_GS_SHOW_WAITDLG
		);
		
	if ( dwRet != WAIT_TIMEOUT )
		return dwRet;
		
	//Show wait dlg
	
	BOOL bComWasInit = FALSE;
	
    HRESULT hr = CoInitializeEx(
		NULL,
		COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE
		);
		
	if ( SUCCEEDED(hr) )
	{
		bComWasInit = TRUE;
	}
	
	COM_DlgWait *pDlgWait = new COM_DlgWait;
	
	if ( pDlgWait != NULL )
	{	
		pDlgWait->Show(
			GetDesktopWindow(),
			LangGetStringW(IDS_GS_WAITDLG_TITLE).c_str(),
			LangGetStringW(IDS_GS_WAITDLG_MESSAGE).c_str(),
			LangGetStringW(IDS_GS_WAITDLG_CANCELMSG).c_str()
			);
			
		while ( GetTickCount() < dwTimeOut )
		{		
			dwRet = WaitForSingleObject(
				hEvent,
				100
				);
				
			if ( dwRet != WAIT_TIMEOUT )
				break;
		
			if ( pDlgWait->IsCancelled() )
			{
				break;
			}		
		}
			
		pDlgWait->EndDialog();
		delete pDlgWait;		
	}
	

	if ( bComWasInit )
	{
		CoUninitialize();
	}

	return dwRet;
}
