/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "GuestCalls.h"
#include <windows.h>
#include "NewClipService.h"
#include "DbgPrintf.h"
#include "GuestServiceConnect.h"
#include "Globals.h"
#include "WinTools.h"
#include "ChkMAlloc.h"

#define DROPPED_FLIST_BUFSIZE (100*1024)


UINT  g_Dbg_WaitInputLastClientId = 0;
UINT  g_Dbg_WaitInputClientCount = 0;
ULONG g_Dbg_WaitLastTick = 0;

UINT
GuestCall_CFH_Begin(
    IN uint32_t ClientId,
    IN uint32_t DropHandle,
    OUT uint32_t *pHandle
    )
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_CFH_BEGIN_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER; //VERR_WRONG_ORDER=-22
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_CFH_BEGIN_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_CFH_BEGIN;
    
    VbglHGCMParmUInt32Set( &Msg.DropHandle, DropHandle );
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmUInt32Set( &Msg.Handle, 0 );    
    
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    DbgPrintf( "VBOX_SHARED_CFH_BEGIN: vbglR3DoIOCtl return: %u(0x%x) Msg.hdr.result: %dL\n", result, result, (ULONG)Msg.hdr.result );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }
    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );
        VbglHGCMParmUInt32Get( &Msg.Handle, pHandle );
    }
    else
    {
        if ( result == VERR_NOT_IMPLEMENTED )
        {
            result = ERROR_INVALID_OPERATION;
        }
		
		if ( result == VERR_INVALID_STATE )
		{
			//Clipboard operation is not allowed
            result = ERROR_RESOURCE_DISABLED;		
		}
		
    }
    
    
    return result;
}

UINT
GuestCall_CFH_GetBuildingListStatus(
    IN  UINT ClientId,
	IN  UINT Handle,
	OUT PCOPY_FILES_BUILDING_LIST_STATUS pCopyFilesBuildingListStatus
	)
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_CFH_GET_BUILDING_LIST_STATUS_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_CFH_GET_BUILDING_LIST_STATUS_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_CFH_GET_BUILDING_LIST_STATUS;
    
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmUInt32Set( &Msg.Handle, Handle );    
    VbglHGCMParmPtrSet( &Msg.StatusInfo, pCopyFilesBuildingListStatus, sizeof(COPY_FILES_BUILDING_LIST_STATUS) );
    
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }
    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );
    }
    else
    {
        result = ERROR_IO_DEVICE;
    }
    
    return result;
}

UINT
GuestCall_CFH_GetCurrentFileName(
    IN  UINT ClientId,
	IN  UINT Handle,
	OUT PCOPY_FILES_GET_CURRENT_FILENAME pCopyFilesGetCurrentFileName
	)
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_CFH_GETCURRENTFILENAME_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_CFH_GETCURRENTFILENAME_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_CFH_GETCURRENTFILENAME;
    
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmUInt32Set( &Msg.Handle, Handle );
    VbglHGCMParmPtrSet( &Msg.FileNameInfo, pCopyFilesGetCurrentFileName, sizeof(COPY_FILES_GET_CURRENT_FILENAME) );
    
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }

    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );

        //Ensure null terminator

        pCopyFilesGetCurrentFileName->wszRelFilePath[ ARRAYSIZE(pCopyFilesGetCurrentFileName->wszRelFilePath) - 1 ] = UNICODE_NULL;
    }
    else
    {
        result = ERROR_IO_DEVICE;
    }
    
    return result;
}

UINT
GuestCall_CFH_GetCurrentStream(
    IN  UINT ClientId,
	IN  UINT Handle,
	IN OUT PVOID pBuffer,
	IN  UINT BufferSize,
	OUT PUINT pBytesCopied,
	OUT PUINT64 pStreamOffset
	)
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_CFH_GET_CURRENTSTREAM_STRUCT Msg;
 
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_CFH_GET_CURRENTSTREAM_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_CFH_GET_CURRENTSTREAM;
    
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmUInt32Set( &Msg.Handle, Handle );
    VbglHGCMParmPtrSet( &Msg.Buffer, pBuffer, BufferSize );
    VbglHGCMParmUInt32Set( &Msg.BytesCopied, 0 );
    VbglHGCMParmUInt64Set( &Msg.StreamOffset, 0 );
    
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }

    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.BytesCopied, pBytesCopied );
        VbglHGCMParmUInt64Get( &Msg.StreamOffset, pStreamOffset );
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );
    }
    else
    {
        result = ERROR_IO_DEVICE;
    }
    
    
    return result;
}

UINT
GuestCall_CFH_MoveToNextItem(
    IN  UINT ClientId,
	IN  UINT Handle,
	UINT64 ItemId
	)
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_MOVE_TO_NEXT_ITEM_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_MOVE_TO_NEXT_ITEM_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_MOVE_TO_NEXT_ITEM;

    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmUInt32Set( &Msg.Handle, Handle );
    VbglHGCMParmUInt64Set( &Msg.NextItemId, ItemId );

    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }

    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );
    }
    else
    {
        result = ERROR_IO_DEVICE;
    }
    
    
    return result;

}


UINT
GuestCall_CFH_End(
	IN  UINT ClientId,
	IN  UINT Handle
	)
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_CFH_END_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_CFH_END_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_CFH_END;
    
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmUInt32Set( &Msg.Handle, Handle );
    
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }
    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );
    }
    else
    {
        result = ERROR_IO_DEVICE;
    }
    
    
    return result;
    
    
    
}

UINT
GuestCall_HSM_NotifyGuestClipboardHasFiles(
    IN  UINT ClientId,
    IN  UINT OSVersion,
    IN  UINT Is64bit
    )
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_HSM_PACKET_GUEST_CLIPBOARD_HAS_FILES;

    DbgPrintf( "GuestCall_HSM_NotifyGuestClipboardHasFiles: OSVersion: 0x%x 64bit: %u\n", OSVersion, Is64bit );

    VbglHGCMParmUInt32Set( &Msg.MajorMinorGuestOSVer, OSVersion );
    VbglHGCMParmUInt32Set( &Msg.Is64bit, Is64bit );
        
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }

    return result; 
}


UINT
GuestCall_HSM_WaitForInputForGuestServer(
    IN  UINT ClientId,
    IN  OUT PVOID pBuffer,
    IN  UINT BufferSize,
    OUT UINT *pOutputSize,
    OUT UINT *pPacketId
    )
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_HSM_PACKET_WAIT_INPUT_FOR_GUEST_SERVER;
    
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, ERROR_GEN_FAILURE );
    VbglHGCMParmPtrSet( &Msg.Buffer, pBuffer, BufferSize );
    VbglHGCMParmUInt32Set( &Msg.OutputBufferSize, 0 );
    VbglHGCMParmUInt32Set( &Msg.PacketId, 0 );
	
	g_Dbg_WaitInputLastClientId = ClientId;
	g_Dbg_WaitInputClientCount++;
	g_Dbg_WaitLastTick = GetTickCount();	
    
	DbgPrintf( "GuestCall_HSM_WaitForInputForGuestServer -> vbglR3DoIOCtl - IN - %u\n", ClientId );
	
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );
	
	DbgPrintf( "GuestCall_HSM_WaitForInputForGuestServer -> vbglR3DoIOCtl - OUT - %u\n", ClientId );	
	
	
	g_Dbg_WaitInputClientCount--;

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }
    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, &result );

        UINT OutputSize = 0;

        VbglHGCMParmUInt32Get( &Msg.OutputBufferSize, &OutputSize );

        if ( OutputSize > BufferSize )
		{
			DbgPrintf( "GuestCall_HSM_WaitForInputForGuestServer: OutputSize > BufferSize\n" );
            return ERROR_GEN_FAILURE;
		}

        *pOutputSize = OutputSize;
        VbglHGCMParmUInt32Get( &Msg.PacketId, pPacketId );

    }
    else
    {
        result = ERROR_IO_DEVICE;

        DbgPrintf( "GuestCall_HSM_WaitForInputForGuestServer: vbglR3DoIOCtl failed: %d\n", result );

    }
    
    
    return result;
}

UINT
GuestCall_HSM_SendOutputFromGuestServer(
    IN  UINT ClientId,
    IN OUT UINT *pErrorCode,
    IN  PVOID pBuffer,
    IN  UINT BufferSize,
    IN  UINT PacketId
    )
{
    UINT result = ERROR_IO_DEVICE;
    VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_HSM_PACKET_SEND_OUTPUT_FROM_SERVER;
    
    VbglHGCMParmUInt32Set( &Msg.ErrorCode, *pErrorCode );
    VbglHGCMParmPtrSet( &Msg.Buffer, pBuffer, BufferSize );  
    VbglHGCMParmUInt32Set( &Msg.PacketId, PacketId );
    
    result = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( result ) )
    {
        result = Msg.hdr.result;
    }
    
    if ( RT_SUCCESS( result ) ) 
    {
        VbglHGCMParmUInt32Get( &Msg.ErrorCode, pErrorCode );
        result = *pErrorCode;
    }
    else
    {
		DbgPrintf( "GuestCall_HSM_SendOutputFromGuestServer: vbglR3DoIOCtl failed: %d\n", result );	
        result = ERROR_IO_DEVICE;
    }
    
    
    return result;

}

VERCHK_STATUS
GuestCall_CheckHostHookVersion(
    PUINT pHostHookVersion,
    PUINT pNativeVboxErrorCode
    )
{
    *pHostHookVersion = 0;
   
    int code = VMInit();
    
    if ( RT_FAILURE( code ) )
    {
        *pNativeVboxErrorCode = code;

        if ( VERR_OPEN_FAILED == code )
        {
            //No tools
            DbgPrintf( "GuestCall_CheckHostHookVersion: Probably, Guest Additions was not installed\n" );
            return VERCHK_TOOLS_ARE_NOT_INSTALLED;
            
        }

        DbgPrintf( "GuestCall_CheckHostHookVersion: VMInit failed: %d\n", code );
        return VERCHK_GEN_FAILURE;
    }

    uint32_t ConnectionId = 0;

    code = CONNECT_FUNCTION( &ConnectionId );

    if ( RT_FAILURE( code ) )
    {
        *pNativeVboxErrorCode = code;
        DbgPrintf( "GuestCall_CheckHostHookVersion: CONNECT_FUNCTION failed: %d\n", code );
        return VERCHK_GEN_FAILURE;
    }

    VERCHK_STATUS ResultCode = VERCHK_SUCCESS;
    *pNativeVboxErrorCode = VINF_SUCCESS;

    VBOX_SHARED_GET_VERSION_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER; //VERR_WRONG_ORDER=-22
    Msg.hdr.u32ClientID =  ConnectionId;
    Msg.hdr.cParms = VBOX_SHARED_GET_VERSION_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_GET_VERSION;
    
    VbglHGCMParmUInt32Set( &Msg.VersionNumber, 0 );

    code = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( code ) )
    {
        code = Msg.hdr.result;
    }
    
    if ( RT_SUCCESS( code ) ) 
    {
        ResultCode = VERCHK_SUCCESS;

        code = VbglHGCMParmUInt32Get( &Msg.VersionNumber, pHostHookVersion );

        if ( !RT_SUCCESS( code ) )
        {
            ResultCode = VERCHK_GEN_FAILURE;
            *pNativeVboxErrorCode = code;
        }
        else if ( *pHostHookVersion != CURRENT_APP_VERSION )
        {
            if ( *pHostHookVersion > CURRENT_APP_VERSION )
            {
                ResultCode = VERCHK_VER_MISMATCH_CURRENT_IS_OLDER;
                DbgPrintf( "GuestCall_CheckHostHookVersion: It need to update VMTransferFiles Guest Module\n" );
            }
            else
            {
                ResultCode = VERCHK_VER_MISMATCH_CURRENT_IS_NEWER;
                DbgPrintf( "GuestCall_CheckHostHookVersion: It need to update VMTransferFiles Host Module\n" );
            }
        }

    }
    else
    {
        *pNativeVboxErrorCode = code;

        if ( code == VERR_NOT_IMPLEMENTED )
        {
            DbgPrintf( "GuestCall_CheckHostHookVersion: Probably, VMTransferFiles Host module was not installed\n" );
            ResultCode = VERCHK_HOSTHOOK_IS_NOT_INSTALLED;
        }
        else
        {
            ResultCode = VERCHK_GEN_FAILURE;
        }

        DbgPrintf( "GuestCall_CheckHostHookVersion: vbglR3DoIOCtl failed: %d\n", code );
    }

    DISCONNECT_FUNCTION( ConnectionId );

    return ResultCode;
}


UINT
VMInit()
{
    static BOOL s_bWasInit = FALSE;
    static UINT s_Code = VINF_SUCCESS;

    if ( s_bWasInit )
        return s_Code;

    RTR3Init();   
    s_Code = VbglR3Init();

    s_bWasInit = TRUE;

    return s_Code;

}

UINT
GuestCall_GetDroppedFileList(
    IN  UINT ClientId OPTIONAL,
	IN  UINT DropHandle,
	OUT vector<wstring> *pFilesList
	)
{
    BOOL bDisconnectRequired = FALSE;
    int code = 0;
    UINT ReturnCode = ERROR_GEN_FAILURE;

    pFilesList->clear();

    if ( ClientId == 0 )
    {
        code = CONNECT_FUNCTION( &ClientId );

        if ( !RT_SUCCESS( code ) )
        {
            return ERROR_GEN_FAILURE;
        }

        bDisconnectRequired = TRUE;
    }

    PUCHAR pBuff = (PUCHAR)MALLOC( DROPPED_FLIST_BUFSIZE );

    if ( NULL == pBuff )
    {
        if ( bDisconnectRequired )
        {
            DISCONNECT_FUNCTION( ClientId );
        }
        
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    VBOX_SHARED_GET_DROPPED_FILES_LIST_STRUCT Msg;
    
    Msg.hdr.result = VERR_WRONG_ORDER;
    Msg.hdr.u32ClientID =  ClientId;
    Msg.hdr.cParms = VBOX_SHARED_GET_DROPPED_FILES_LIST_CPARAMS;
    Msg.hdr.u32Function = VBOX_SHARED_GET_DROPPED_FILES_LIST;

    VbglHGCMParmUInt32Set( &Msg.DropHandle, DropHandle );
    VbglHGCMParmPtrSet( &Msg.FileListBuffer, pBuff, DROPPED_FLIST_BUFSIZE );
    VbglHGCMParmUInt32Set( &Msg.FileListBufferWriteln, 0 );

    code = vbglR3DoIOCtl( VBOXGUEST_IOCTL_HGCM_CALL( sizeof(Msg) ), &Msg, sizeof(Msg) );

    if ( RT_SUCCESS( code ) )
    {
        code = Msg.hdr.result;
    }

    ReturnCode = ERROR_GEN_FAILURE;
    
    if ( RT_SUCCESS( code ) ) 
    {
        UINT FileListBufferWriteln = 0;

        VbglHGCMParmUInt32Get( &Msg.FileListBufferWriteln, &FileListBufferWriteln );

        if ( FileListBufferWriteln > 0 && FileListBufferWriteln <= DROPPED_FLIST_BUFSIZE )
        {
            if ( BuffToStringListW( pFilesList, pBuff, (ULONG)FileListBufferWriteln ) )
            {
                ReturnCode = ERROR_SUCCESS;
            }
            else
            {
                DbgPrintf( "BuffToStringListW failed\n" );
            }
        }
    }

    FREE( pBuff );

    if ( bDisconnectRequired )
    {
        DISCONNECT_FUNCTION( ClientId );
    }

    return ReturnCode;
}

VOID
Dbg_GuestCall_DumpInfo()
{
	DbgPrintf( "InpLastClientId=%u Clients=%u Tick=%u Seconds=%u\n", g_Dbg_WaitInputLastClientId, g_Dbg_WaitInputClientCount, g_Dbg_WaitLastTick, (GetTickCount()-g_Dbg_WaitLastTick)/1000 );
}

