/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "PasteFilesFromHost.h"

#include "vboxhdrs.h"
#include <windows.h>

#include "NewClipService.h"
#include "GuestHostStructs.h"
#include "CopyFilesHandler.h"
#include "WinTools.h"
#include "DbgPrintf.h"
#include "LangTools.h"
#include "ConfirmDlg.h"
#include "DllMain.h"
#include "GuestServiceConnect.h"
#include "GuestCalls.h"
#include "Globals.h"
#include "Settings.h"


extern
string
GetVerChkMsg(
    VERCHK_STATUS VerChkCode,
    UINT VerChkNativeErrCode
    );



VOID
PasteFilesFromHost(
    LPCWSTR lpszDestFolderName
    )
{
    static BOOL bWasInit = FALSE;
    int code = 0;

    UINT HostHookVersion = 0;
    UINT NativeVboxErrorCode = 0;

    VERCHK_STATUS VerChkStatus = GuestCall_CheckHostHookVersion( &HostHookVersion, &NativeVboxErrorCode );

    if ( VerChkStatus != VERCHK_SUCCESS )
    {
        MessageBox( NULL, GetVerChkMsg( VerChkStatus, NativeVboxErrorCode ).c_str(), APP_NAMEA, MB_SETFOREGROUND | MB_APPLMODAL | MB_ICONWARNING  );
        return;
    }


    if ( !bWasInit )
    {       
        code = VMInit();
        
        if ( RT_FAILURE( code ) )
        {
            DbgPrintf( "VMInit failed: %d (0x%x)\n", code, code );

            if ( VERR_OPEN_FAILED == code )
            {
                LangErrorMessageBox( IDS_INITF_NOTOOLS );
            }
            else
            {
                LangErrorMessageBoxFmt( IDS_FUNC_FAILED_LONG_FMT, L"VMInit", code );
            }

            return ;
        }

        bWasInit = TRUE;
    } 


    uint32_t ConnectionId = 0;
    
    code = CONNECT_FUNCTION( &ConnectionId );

    if ( RT_SUCCESS( code ) )
    {
        CGuestFromHostCopierHelper CopyingHelper;

        ULONG uRet = ProceeedCopyingFiles( &CopyingHelper, ConnectionId, 0, lpszDestFolderName );
		
		if ( uRet > 0 && Settings_BeepAfterCopyingFiles() )
		{
			MessageBeep( MB_OK );
		}
		
		
        DISCONNECT_FUNCTION( ConnectionId );
    }
    else
    {
        LangErrorMessageBoxFmt( IDS_FUNC_FAILED_LONG_FMT, L"CONNECT_FUNCTION", code );
    }
}


EXTERN_C
BOOL
CALLBACK
TryConnectToHostOS()
{
    int code = VMInit();
    
    if ( RT_FAILURE( code ) )
        return FALSE;

    return TRUE;

}
