/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "ExplorerClassFactory.h"
#include "ExplorerContextMenu.h"
#include "DbgPrintf.h"
#include "guid.h"

STDMETHODIMP_(ULONG) 
CExplorerClassFactory::AddRef()
{
    return ++m_cRef;
}

STDMETHODIMP_(ULONG) 
CExplorerClassFactory::Release()
{
    ULONG cRef = --m_cRef;

    if ( cRef == 0 )
        delete this;

    return cRef;
}

STDMETHODIMP
CExplorerClassFactory::QueryInterface(
    REFIID riid,
    PVOID *ppv
    )
{
    //DbgPrintf( "CExplorerClassFactory::QueryInterface\n" );

    if ( IsEqualIID( riid, IID_IUnknown ) ||
         IsEqualIID( riid, IID_IClassFactory )
        )
    {
        AddRef();
        *ppv = this;
        return NOERROR;
    }
    else
    {
        *ppv = NULL;
        return E_NOINTERFACE;
    }

}


STDMETHODIMP
CExplorerClassFactory::CreateInstance(
    IUnknown *punkOuter,
    REFIID iid,
    PVOID *ppv
    )
{
    *ppv = NULL;

    //DbgPrintf( "CExplorerClassFactory::CreateInstance\n" );

    if ( NULL != punkOuter )
        return CLASS_E_NOAGGREGATION;

    CExplorerContextMenu *pExplorerContextMenu = new CExplorerContextMenu;

    if ( pExplorerContextMenu == NULL )
        return E_OUTOFMEMORY;

    HRESULT hr = pExplorerContextMenu->QueryInterface( iid, ppv );

    if ( hr != NOERROR )
    {
        DbgPrintf( "pExplorerContextMenu->QueryInterface failed: %x\n", hr );
        delete pExplorerContextMenu;
        *ppv = NULL;
    }

    return hr;
}

