
/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "CopyFilesHandler.h"
#include "WinTools.h"
#include <vector>
#include <string>
#include <map>
#include "DbgPrintf.h"
#include "CriticalSection.h"
#include "SecureFuncs.h"
#include "ChkMAlloc.h"
#include <strsafe.h>
using namespace std;

//=====================================================================


typedef struct CFH_FILE_ITEM_INFO
{
	wstring  FileName;
	wstring  RelFileName;
	DWORD    Attributes;
	UINT64  FileSize;
    FILETIME LastWriteTime;
    UINT64   ItemId;

}CFH_FILE_ITEM_INFO;

typedef struct COPY_FILES_OBJECT
{
	volatile LONG		    RefCount;
	UINT		    ClientId;
	UINT		    Handle;
	volatile UINT		    Stage;
	vector<wstring> ClipboardFileList;
	vector<CFH_FILE_ITEM_INFO> FullFileList;
	CFH_FILE_ITEM_INFO		   CurrentFile;
	HANDLE					   CurrentFileHandle;
	UINT64					   CurrentFileOffset;

	volatile UINT		   FileCount;
	volatile UINT		   FolderCount;
	volatile UINT64       AllStreamSize;

	volatile HANDLE		    hWorkerThread;
	volatile LONG           AbortFlag;
    UINT64                  NextCreatedItemId;


} COPY_FILES_OBJECT, *PCOPY_FILES_OBJECT;

typedef struct CFH_BUILDLIST_PARAMS
{
	PCOPY_FILES_OBJECT pObject;
	wstring ParentDir;

} CFH_BUILDLIST_PARAMS, *PCFH_BUILDLIST_PARAMS;


//=====================================================================

BOOL
CALLBACK
StandardGetCopiedFileListHandler(
    UINT DropHandle,
    vector<wstring> *pClipboardFileList
    );


DWORD 
WINAPI 
CFH_BuildFileListWorker(
	LPVOID lpParameter
	);

BOOL
__stdcall
CFH_BuildList_EnumDirProcW(
	PVOID pContext,
	LPCWSTR lpszFileName,
	DWORD dwAttr,
	UINT64 FileSize,
    FILETIME *pLastWriteTime
	);

VOID
ReleaseCopyFilesObject(
   PCOPY_FILES_OBJECT pNewObject
   );



//=====================================================================

static UINT s_LastFreeHandle = 1;
map<UINT, PCOPY_FILES_OBJECT> g_CopyFilesMap;
CCriticalSection g_CopyFilesMapLock;
TFILELISTTOCOPYHANDLER g_FileListToCopyHandler = StandardGetCopiedFileListHandler;

//=====================================================================

UINT
CFH_ReferenceObjectByHandle( UINT Handle, PCOPY_FILES_OBJECT *ppObject )
{
    UINT ErrorCode = ERROR_SUCCESS;

    g_CopyFilesMapLock.Lock();

		map<UINT, PCOPY_FILES_OBJECT>::iterator iter = g_CopyFilesMap.find(
			Handle
			);

    if ( iter != g_CopyFilesMap.end() )
    {
        *ppObject = iter->second;
        InterlockedIncrement( &iter->second->RefCount );
    }
    else
    {
        ErrorCode = ERROR_INVALID_PARAMETER;
    }

    g_CopyFilesMapLock.Unlock();

    return ErrorCode;
}

UINT
CFH_ReferenceObjectByHandleSafe( UINT Handle, UINT ClientId, PCOPY_FILES_OBJECT *ppObject )
{
    UINT ErrorCode = ERROR_SUCCESS;

    g_CopyFilesMapLock.Lock();

		map<UINT, PCOPY_FILES_OBJECT>::iterator iter = g_CopyFilesMap.find(
			Handle
			);

    if ( iter != g_CopyFilesMap.end() )
    {
        if ( iter->second->ClientId == ClientId )
        {
            *ppObject = iter->second;
            InterlockedIncrement( &iter->second->RefCount );
        }
        else
        {
            ErrorCode = ERROR_ACCESS_DENIED;
        }
    }
    else
    {
        ErrorCode = ERROR_INVALID_PARAMETER;
    }

    g_CopyFilesMapLock.Unlock();

    return ErrorCode;
}


VOID
ReleaseCopyFilesObject(
   PCOPY_FILES_OBJECT pObject
   )
{

	if ( 0 == InterlockedDecrement( &pObject->RefCount ) )
	{
        g_CopyFilesMapLock.Lock();

        if ( pObject->RefCount != 0 )
        {
            g_CopyFilesMapLock.Unlock();
            return;
        }

		map<UINT, PCOPY_FILES_OBJECT>::iterator iter = g_CopyFilesMap.find(
			pObject->Handle
			);

		if ( iter != g_CopyFilesMap.end() )
		{
			g_CopyFilesMap.erase( iter );
		}

        if ( pObject->CurrentFileHandle != INVALID_HANDLE_VALUE )
        {
            CloseHandle( pObject->CurrentFileHandle );
            pObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
        }

		delete pObject;

        g_CopyFilesMapLock.Unlock();

	}


}

UINT
CFH_Begin(
    IN  UINT ClientId,
    IN  UINT DropHandle,
	OUT PUINT pHandle
	)
{
	vector<wstring> ClipboardFileList;

	if ( !g_FileListToCopyHandler( DropHandle, &ClipboardFileList ) )
	{
		return ERROR_NO_DATA;
	}

	PCOPY_FILES_OBJECT pNewObject = new COPY_FILES_OBJECT;

	if ( pNewObject == NULL )
		return ERROR_NO_SYSTEM_RESOURCES;

	//for new thread
	pNewObject->RefCount = 2;
	pNewObject->ClientId = ClientId;
	pNewObject->Handle = s_LastFreeHandle++;
	pNewObject->Stage = COPY_FILES_STAGE_BUILDING_LIST;
	pNewObject->FileCount = 0;
	pNewObject->FolderCount = 0;
	pNewObject->AllStreamSize = 0;
	pNewObject->hWorkerThread = NULL;
	pNewObject->AbortFlag = 0;
	pNewObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
	pNewObject->CurrentFileOffset = 0;
    pNewObject->NextCreatedItemId = 0;

	pNewObject->ClipboardFileList = ClipboardFileList;

    g_CopyFilesMapLock.Lock();
	g_CopyFilesMap[ pNewObject->Handle ] = pNewObject;
    g_CopyFilesMapLock.Unlock();

	DWORD dwTid = 0;

	pNewObject->hWorkerThread = CreateThread(
		NULL,
		0,
		CFH_BuildFileListWorker,
		pNewObject,
		0,
		&dwTid
		);

	if ( pNewObject->hWorkerThread == NULL )
	{
		ReleaseCopyFilesObject( pNewObject );
		ReleaseCopyFilesObject( pNewObject );
		return ERROR_NO_SYSTEM_RESOURCES;
	}


	*pHandle = pNewObject->Handle;
	return ERROR_SUCCESS;
}



DWORD 
WINAPI 
CFH_BuildFileListWorker(
	LPVOID lpParameter
	)
{
	PCOPY_FILES_OBJECT pObject = (PCOPY_FILES_OBJECT)lpParameter;

	for (
		vector<wstring>::iterator iter = pObject->ClipboardFileList.begin();
		iter != pObject->ClipboardFileList.end() && pObject->AbortFlag == 0;
		iter++
		)
	{
		UINT64 FileSize = 0;
		DWORD FileAttr = 0;

        FILETIME LastWriteTime;

        RtlZeroMemory( &LastWriteTime, sizeof(LastWriteTime) );

		BOOL bRet = GetFileInfoW(
			iter->c_str(),
			&FileSize,
			&FileAttr,
            &LastWriteTime
			);

		if ( !bRet )
		{
			DbgPrintf( "GetFileInfoW failed\n" );
			continue;
		}

		CFH_FILE_ITEM_INFO NewItem;

		NewItem.FileName = *iter;
		NewItem.Attributes = FileAttr;
		NewItem.FileSize = FileSize;
        NewItem.LastWriteTime = LastWriteTime;
		NewItem.RelFileName = GetBaseFileNameW(
			NewItem.FileName.c_str()
			);

        NewItem.ItemId = pObject->NextCreatedItemId++;

		pObject->FullFileList.push_back( NewItem );

		if ( NewItem.Attributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			pObject->FolderCount++;

			CFH_BUILDLIST_PARAMS Params;			

			LPCWSTR lpwszParentDirEnd = wcsrchr( NewItem.FileName.c_str(), L'\\' );

			if ( lpwszParentDirEnd )
			{
				Params.pObject = pObject;

				Params.ParentDir.assign(
					NewItem.FileName.c_str(),
					lpwszParentDirEnd - NewItem.FileName.c_str()
					);

				EnumerateDirectoriesW(
					NewItem.FileName.c_str(),
					&Params,
					CFH_BuildList_EnumDirProcW
					);
			}

		}
		else
		{
			pObject->FileCount++;
			pObject->AllStreamSize += FileSize;
		}

	}

	//=====================

	if ( pObject->Stage == COPY_FILES_STAGE_BUILDING_LIST )
	{
		pObject->Stage = COPY_FILES_STAGE_GET_CURRENT_FILENAME;
	}

	CloseHandle( pObject->hWorkerThread );
	pObject->hWorkerThread = NULL;
	ReleaseCopyFilesObject( pObject );

	return 0;
}

BOOL
__stdcall 
CFH_BuildList_EnumDirProcW(
	PVOID pContext,
	LPCWSTR lpszFileName,
	DWORD dwAttr,
	UINT64 FileSize,
    FILETIME *pLastWriteTime
	)
{
	PCFH_BUILDLIST_PARAMS pParams = (PCFH_BUILDLIST_PARAMS)pContext;
	PCOPY_FILES_OBJECT pObject = pParams->pObject;

	if ( pObject->AbortFlag )
		return FALSE;

	if ( 0 != _wcsnicmp( lpszFileName, pParams->ParentDir.c_str(), pParams->ParentDir.length() ) )
	{
		DbgPrintf( "*****CFH_BuildList_EnumDirProcW: BUG,BUG\n" );
		return TRUE; //continue checking
	}

	if ( dwAttr & FILE_ATTRIBUTE_DIRECTORY )
	{
		pObject->FolderCount++;
	}
	else
	{
		pObject->FileCount++;
		pObject->AllStreamSize += FileSize;
	}

	CFH_FILE_ITEM_INFO NewItem;

	NewItem.FileName = lpszFileName;
	NewItem.Attributes = dwAttr;
	NewItem.FileSize = FileSize;

    //FIXED
    NewItem.LastWriteTime = *pLastWriteTime;
    NewItem.ItemId = pObject->NextCreatedItemId++;


	LPCWSTR lpszRelPath = lpszFileName + pParams->ParentDir.length();

	if ( lpszRelPath[0] == L'\\' )
		lpszRelPath++;

	NewItem.RelFileName = lpszRelPath;

	pObject->FullFileList.push_back( NewItem );

	return TRUE;
}


UINT
CFH_End(
	IN  UINT ClientId,
	IN  UINT Handle
	)
{
    UINT ResultCode = E_FAIL;
    PCOPY_FILES_OBJECT pObject = NULL;

    ResultCode = CFH_ReferenceObjectByHandleSafe( Handle, ClientId, &pObject );

    if ( ResultCode != ERROR_SUCCESS )
    {
        return ResultCode;
    }

	UINT CapturedStage = pObject->Stage;                     
	pObject->Stage = COPY_FILES_STAGE_RELEASE_OBJ;

	pObject->AbortFlag = TRUE;

	if ( pObject->CurrentFileHandle != INVALID_HANDLE_VALUE )
	{
		HANDLE CapturedHandle = pObject->CurrentFileHandle;
		pObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
		CloseHandle( CapturedHandle );
	}

    //release it twice because we already added refs

	ReleaseCopyFilesObject( pObject );

    if ( CapturedStage != COPY_FILES_STAGE_RELEASE_OBJ )
    {
        ReleaseCopyFilesObject( pObject );
    }

	return ERROR_SUCCESS;
}

/*
VOID
CFH_Dump(
	IN  UINT Handle
	)
{
	printf( "CFH_Dump for %u handle\n", Handle );

	map<UINT, PCOPY_FILES_OBJECT>::iterator iter = g_CopyFilesMap.find( Handle );

	if ( g_CopyFilesMap.end() == iter )
	{
		printf( "CFH: Invalid handle\n" );
		return ;
	}

	PCOPY_FILES_OBJECT pObject = iter->second;

	printf( "ClientId: %u Stage: %u\n", pObject->ClientId, pObject->Stage );
	printf( "FileCount: %u FolderCount: %u AllStreamSize: %I64u FullFileList: %u\n", pObject->FileCount, pObject->FolderCount, pObject->AllStreamSize, pObject->FullFileList.size() );
		

	for (
		vector<CFH_FILE_ITEM_INFO>::iterator iter = pObject->FullFileList.begin();
		iter != pObject->FullFileList.end();
		iter++
		)
	{
		printf( "%ws - %ws\n", iter->RelFileName.c_str(), iter->FileName.c_str() ); 
	}
}
*/

UINT
CFH_GetBuildingListStatus(
    IN  UINT ClientId,
	IN  UINT Handle,
	OUT PCOPY_FILES_BUILDING_LIST_STATUS pCopyFilesBuildingListStatus
	)
{

	if ( pCopyFilesBuildingListStatus->StructSize != sizeof(COPY_FILES_BUILDING_LIST_STATUS) )
		return ERROR_INVALID_PARAMETER;

    PCOPY_FILES_OBJECT pObject = NULL;
    UINT ResultCode = CFH_ReferenceObjectByHandleSafe( Handle, ClientId, &pObject );

    if ( ResultCode != ERROR_SUCCESS )
        return ResultCode;

	if ( pObject->Stage == COPY_FILES_STAGE_BUILDING_LIST )
	{
		ResultCode = ERROR_IO_PENDING;
	}

	pCopyFilesBuildingListStatus->AllStreamSize = pObject->AllStreamSize; 
	pCopyFilesBuildingListStatus->FileCount = pObject->FileCount;
	pCopyFilesBuildingListStatus->FolderCount = pObject->FolderCount;

    ReleaseCopyFilesObject( pObject );
	return ResultCode;

}

UINT
CFH_GetCurrentFileName(
    IN  UINT ClientId,
	IN  UINT Handle,
	OUT PCOPY_FILES_GET_CURRENT_FILENAME pCopyFilesGetCurrentFileName
	)
{
	if ( pCopyFilesGetCurrentFileName->StructSize != sizeof(COPY_FILES_GET_CURRENT_FILENAME) )
		return ERROR_INVALID_PARAMETER;

	PCOPY_FILES_OBJECT pObject = NULL;
    UINT ResultCode = CFH_ReferenceObjectByHandleSafe( Handle, ClientId, &pObject );

    if ( ResultCode != ERROR_SUCCESS )
        return ResultCode;

	if ( pObject->Stage != COPY_FILES_STAGE_GET_CURRENT_FILENAME )
	{
		if ( pObject->Stage == COPY_FILES_STAGE_RELEASE_OBJ || COPY_FILES_STAGE_DONE == pObject->Stage )
        {
			ResultCode = ERROR_NO_MORE_FILES;
        }
        else
        {
		    ResultCode = ERROR_INVALID_OPERATION;
        }

        ReleaseCopyFilesObject( pObject );
        return ResultCode;
	}

	if ( pObject->FullFileList.size() == 0 )
	{
		pObject->Stage = COPY_FILES_STAGE_DONE;
        ReleaseCopyFilesObject( pObject );
		return ERROR_NO_MORE_FILES;
	}

	pObject->CurrentFile = *pObject->FullFileList.begin();
	pObject->FullFileList.erase( pObject->FullFileList.begin() );

	StringCchCopyW(
		pCopyFilesGetCurrentFileName->wszRelFilePath,
		ARRAYSIZE(pCopyFilesGetCurrentFileName->wszRelFilePath),
		pObject->CurrentFile.RelFileName.c_str()
		);


	pCopyFilesGetCurrentFileName->FileSize = pObject->CurrentFile.FileSize;
	pCopyFilesGetCurrentFileName->FileAttr = pObject->CurrentFile.Attributes;
    pCopyFilesGetCurrentFileName->FileLastWriteTime = pObject->CurrentFile.LastWriteTime;
    pCopyFilesGetCurrentFileName->ItemId = pObject->CurrentFile.ItemId;

	pCopyFilesGetCurrentFileName->bHasStream = (0 != pCopyFilesGetCurrentFileName->FileSize);

	if ( pCopyFilesGetCurrentFileName->bHasStream )
	{
		pObject->Stage = COPY_FILES_STAGE_GET_CURRENT_STREAM;

		if ( pObject->CurrentFileHandle != INVALID_HANDLE_VALUE )
		{
			CloseHandle( pObject->CurrentFileHandle );
			pObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
			pObject->CurrentFileOffset = 0;
		}

	}

    ReleaseCopyFilesObject( pObject );
	return ERROR_SUCCESS;
}

UINT
CFH_GetCurrentStream(
    IN  UINT ClientId,
	IN  UINT Handle,
	IN OUT PVOID pBuffer,
	IN  UINT BufferSize,
	OUT PUINT pBytesCopied,
	OUT PUINT64 pStreamOffset
	)
{
	PCOPY_FILES_OBJECT pObject = NULL;
    UINT ResultCode = CFH_ReferenceObjectByHandleSafe( Handle, ClientId, &pObject );
    DWORD dwGle = ERROR_OPEN_FAILED;

    if ( ResultCode != ERROR_SUCCESS )
        return ResultCode;

	if ( pObject->Stage != COPY_FILES_STAGE_GET_CURRENT_STREAM )
	{
		if ( pObject->Stage == COPY_FILES_STAGE_GET_CURRENT_FILENAME ||
			 pObject->Stage == COPY_FILES_STAGE_DONE
			)
		{
            ReleaseCopyFilesObject( pObject );
			return ERROR_NO_DATA;
		}

        ReleaseCopyFilesObject( pObject );
		return ERROR_INVALID_OPERATION;
	}

	if ( pObject->CurrentFileHandle == INVALID_HANDLE_VALUE )
	{
		pObject->CurrentFileOffset = 0;

		pObject->CurrentFileHandle = CreateFileW(
			pObject->CurrentFile.FileName.c_str(),
			GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL
			);

        dwGle = GetLastError();

		if ( pObject->CurrentFileHandle == INVALID_HANDLE_VALUE )
		{
            ReleaseCopyFilesObject( pObject );
			return dwGle;
		}
	}

	ULONG ReadBytes = 0;
	BOOL bRet = FALSE;

	bRet = ReadFile(
		pObject->CurrentFileHandle,
		pBuffer,
		BufferSize,
		&ReadBytes,
		NULL
		);

	if ( ( !bRet && ERROR_HANDLE_EOF == GetLastError() ) ||
		 ( bRet && ReadBytes == 0 )
		)
	{
		CloseHandle( pObject->CurrentFileHandle );
		pObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
		pObject->CurrentFileOffset = 0;
		pObject->Stage = COPY_FILES_STAGE_GET_CURRENT_FILENAME;

        ReleaseCopyFilesObject( pObject );
		return ERROR_HANDLE_EOF;		
	}

	if ( !bRet )
	{
		CloseHandle( pObject->CurrentFileHandle );
		pObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
		pObject->CurrentFileOffset = 0;
		pObject->Stage = COPY_FILES_STAGE_GET_CURRENT_FILENAME;

        ReleaseCopyFilesObject( pObject );
		return ERROR_READ_FAULT;
	}


	pObject->CurrentFileOffset = pObject->CurrentFileOffset + ReadBytes;
	*pStreamOffset = pObject->CurrentFileOffset;
	*pBytesCopied = ReadBytes;

    ReleaseCopyFilesObject( pObject );
	return ERROR_SUCCESS;

}

UINT
CFH_ClientDisconnected(
	IN  UINT ClientId
	)
{
    g_CopyFilesMapLock.Lock();

    for (
        map<UINT, PCOPY_FILES_OBJECT>::iterator iter = g_CopyFilesMap.begin();
        iter != g_CopyFilesMap.end();
        )
    {
        map<UINT, PCOPY_FILES_OBJECT>::iterator next_iter = iter;
        next_iter++;
            
        if ( ClientId == iter->second->ClientId )
        {
            UINT CapturedStage = iter->second->Stage;
            iter->second->Stage = COPY_FILES_STAGE_RELEASE_OBJ;
           
            if ( CapturedStage != COPY_FILES_STAGE_RELEASE_OBJ )
            {
                ReleaseCopyFilesObject( iter->second );
            }
        }
        
        iter = next_iter;
    }

    g_CopyFilesMapLock.Unlock();    
    return ERROR_SUCCESS;
}

UINT
CFH_MoveToNextItem(
    IN  UINT ClientId,
	IN  UINT Handle,
	UINT64 ItemId
	)
{
	PCOPY_FILES_OBJECT pObject = NULL;
    UINT ResultCode = CFH_ReferenceObjectByHandleSafe( Handle, ClientId, &pObject );

    if ( ResultCode != ERROR_SUCCESS )
        return ResultCode;

    if ( pObject->Stage == COPY_FILES_STAGE_DONE )
    {
        ReleaseCopyFilesObject( pObject );
		return ERROR_NO_MORE_FILES;
    }

	if ( pObject->Stage != COPY_FILES_STAGE_GET_CURRENT_STREAM &&
         pObject->Stage != COPY_FILES_STAGE_GET_CURRENT_FILENAME
        )
	{
        ReleaseCopyFilesObject( pObject );
		return ERROR_INVALID_OPERATION;
	}

    if ( pObject->CurrentFile.ItemId >= ItemId )
    {
        ReleaseCopyFilesObject( pObject );
		return ERROR_INVALID_OPERATION;
    }

    if ( INVALID_HANDLE_VALUE != pObject->CurrentFileHandle )
    {
        CloseHandle(pObject->CurrentFileHandle);
		pObject->CurrentFileHandle = INVALID_HANDLE_VALUE;
    }

    pObject->CurrentFileOffset = 0;
    pObject->Stage = COPY_FILES_STAGE_GET_CURRENT_FILENAME;

	while ( pObject->FullFileList.size() != 0 )    
	{
        if ( pObject->FullFileList.begin()->ItemId == ItemId )
        {
            //copying info is on COPY_FILES_STAGE_GET_CURRENT_FILENAME stage
            ReleaseCopyFilesObject( pObject );
            return ERROR_SUCCESS;
        }

    	pObject->CurrentFile = *pObject->FullFileList.begin();
    	pObject->FullFileList.erase( pObject->FullFileList.begin() );


	}

    pObject->Stage = COPY_FILES_STAGE_DONE;
    ReleaseCopyFilesObject( pObject );
    return ERROR_NO_MORE_FILES;
}


BOOL
CALLBACK
StandardGetCopiedFileListHandler(
    UINT DropHandle,
    vector<wstring> *pClipboardFileList
    )
{
    if ( DropHandle != 0 )
        return FALSE;

    return ClipboardGetCopiedFileList( pClipboardFileList );
}

TFILELISTTOCOPYHANDLER
SetGetCopiedFileListHandler( TFILELISTTOCOPYHANDLER pNewHandler )
{
    TFILELISTTOCOPYHANDLER Result = g_FileListToCopyHandler;
    g_FileListToCopyHandler = pNewHandler;

    return Result;
}
