/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "FileOperationsProgressDlgs.h"
#include <stdio.h>
#include "LangTools.h"
#include "SecureFuncs.h"
#include "ChkMAlloc.h"
#include <strsafe.h>


#ifndef PROGDLG_MARQUEEPROGRESS
#define PROGDLG_MARQUEEPROGRESS 0x00000020
#endif

BOOL
IsVistaOrLater();


COM_CPrepareFileCopyingDlg::COM_CPrepareFileCopyingDlg()
{
	m_ProgressDlg = NULL;
	m_InitFailed = FALSE;

	m_Shell32Dll = LoadLibrary( "shell32.dll" );
}

COM_CPrepareFileCopyingDlg::~COM_CPrepareFileCopyingDlg()
{
    

	if ( m_ProgressDlg )
	{
        EndDialog();
		m_ProgressDlg->Release();
		m_ProgressDlg = NULL;
	}

	if ( m_Shell32Dll )
	{
		FreeLibrary( m_Shell32Dll );
		m_Shell32Dll = NULL;
	}

}

BOOL
IsVistaOrLater()
{
    OSVERSIONINFOEX OsVersionInfoEx;

    RtlZeroMemory( &OsVersionInfoEx, sizeof(OsVersionInfoEx) );
    OsVersionInfoEx.dwOSVersionInfoSize = sizeof(OsVersionInfoEx);

    GetVersionEx( (OSVERSIONINFO *)&OsVersionInfoEx );

    if ( OsVersionInfoEx.dwMajorVersion >= 6 )
        return TRUE;
	else
		return FALSE;


}

VOID
COM_CPrepareFileCopyingDlg::ShowProgress(
	HWND hWndParent,
    ULONG FileCount,
    ULONG FolderCount,
	ULONG64 DiscoveredItemsSize,
	LPCWSTR lpszDestFolder
    )
{
	if ( m_InitFailed )
		return;

	if ( m_ProgressDlg == NULL )
	{
		HRESULT hr = CoCreateInstance(
			CLSID_ProgressDialog, 
			NULL, 
			CLSCTX_INPROC_SERVER, 
			IID_IProgressDialog, 
			(void **)&m_ProgressDlg
			);

		if ( FAILED(hr) )
		{
			m_InitFailed = TRUE;
			m_ProgressDlg = NULL;
			return;
		}

		if ( m_Shell32Dll )
		{
			m_ProgressDlg->SetAnimation(
				m_Shell32Dll, 
				160
				);
		}

		m_ProgressDlg->SetTitle( LangGetStringW( IDS_PREPARING_TO_COPY_TITLE ).c_str() );
		m_ProgressDlg->SetLine( 1, LangGetStringW( IDS_PREPARING_TO_COPY_TITLE ).c_str(), FALSE, NULL );
		m_ProgressDlg->SetCancelMsg( LangGetStringW( IDS_PREPARING_TO_COPY_CANCEL_MSG ).c_str(), NULL );

        m_ProgressDlg->Timer( PDTIMER_RESET, NULL );

		DWORD dwFlags = PROGDLG_NOTIME;

		if ( IsVistaOrLater() )
		{
			dwFlags |= PROGDLG_MARQUEEPROGRESS;
		}
		else
		{
			dwFlags |= PROGDLG_NOPROGRESSBAR;
		}

        m_ProgressDlg->StartProgressDialog(
				hWndParent, 
				NULL, 
				dwFlags,
				NULL);


	}

	WCHAR wszLine[ 300 ] = L"";

	StringCchPrintfW(
		wszLine,
		ARRAYSIZE(wszLine),		
		LangGetStringW( IDS_PREPARING_TO_COPY_DEST_DIR_FMT ).c_str(),
		lpszDestFolder
		); 

	m_ProgressDlg->SetLine( 2, wszLine, FALSE, NULL);


	StringCchPrintfW( 
		wszLine, 
		ARRAYSIZE(wszLine), 
		LangGetStringW( IDS_ANALYZED_ITEMS_FMT ).c_str(), 
		FileCount + FolderCount
		); 

	m_ProgressDlg->SetLine( 3, wszLine, FALSE, NULL);

}

VOID
COM_CPrepareFileCopyingDlg::EndDialog()
{
	if ( m_ProgressDlg )
	{
		m_ProgressDlg->StopProgressDialog();
	}
}


BOOL
COM_CPrepareFileCopyingDlg::IsCancelled()
{
	if ( NULL == m_ProgressDlg )
		return FALSE;

	return m_ProgressDlg->HasUserCancelled();
}



COM_CFileCopyingProgressDlg::COM_CFileCopyingProgressDlg()
{
	m_ProgressDlg = NULL;
	m_InitFailed = FALSE;

	m_Shell32Dll = LoadLibrary( "shell32.dll" );

}

COM_CFileCopyingProgressDlg::~COM_CFileCopyingProgressDlg()
{
	if ( m_Shell32Dll )
	{
		FreeLibrary( m_Shell32Dll );
		m_Shell32Dll = NULL;
	}

	if ( m_ProgressDlg )
	{
		m_ProgressDlg->Release();
		m_ProgressDlg = NULL;
	}

}


VOID
COM_CFileCopyingProgressDlg::ShowProgress(
	HWND hWndParent,
	ULONG   AllFilesAndFoldersCount,
	ULONG64 ProcessedItemsSize,
	ULONG64 AllItemsSize,
	LPCWSTR lpszDestFile
    )
{
	WCHAR wszStr[ 300 ] = L"";

	if ( m_InitFailed )
		return;

	if ( m_ProgressDlg == NULL )
	{
		HRESULT hr = CoCreateInstance(
			CLSID_ProgressDialog, 
			NULL, 
			CLSCTX_INPROC_SERVER, 
			IID_IProgressDialog, 
			(void **)&m_ProgressDlg
			);

		if ( FAILED(hr) )
		{
			m_InitFailed = TRUE;
			m_ProgressDlg = NULL;
			return;
		}

		if ( m_Shell32Dll )
		{
			m_ProgressDlg->SetAnimation(
				m_Shell32Dll, 
				160
				);
		}

		StringCchPrintfW(
			wszStr,
			ARRAYSIZE(wszStr),
			LangGetStringW( IDS_COPY_FILES_TITLE_FMT ).c_str(),
			AllFilesAndFoldersCount
			); 


		m_ProgressDlg->SetTitle( wszStr );
		m_ProgressDlg->SetLine( 1, wszStr, FALSE, NULL );
		m_ProgressDlg->SetCancelMsg( LangGetStringW( IDS_PREPARING_TO_COPY_CANCEL_MSG ).c_str(), NULL );

        m_ProgressDlg->Timer( PDTIMER_RESET, NULL );

		DWORD dwFlags = PROGDLG_AUTOTIME;

        m_ProgressDlg->StartProgressDialog(
				hWndParent, 
				NULL, 
				dwFlags,
				NULL);


	}

	m_ProgressDlg->SetProgress64( ProcessedItemsSize, AllItemsSize );

	StringCchPrintfW(
		wszStr,
		ARRAYSIZE(wszStr),
		LangGetStringW( IDS_COPY_FILES_DST_FILE ).c_str(),
		lpszDestFile
		); 


	m_ProgressDlg->SetLine( 2, wszStr, FALSE, NULL );

	StringCchPrintfW(
		wszStr,
		ARRAYSIZE(wszStr),
		LangGetStringW( IDS_ANALYZED_ITEMS_FMT ).c_str(),
		AllFilesAndFoldersCount
		);

	m_ProgressDlg->SetLine( 3, wszStr, FALSE, NULL );

}

VOID
COM_CFileCopyingProgressDlg::EndDialog()
{
	if ( m_ProgressDlg )
	{
		m_ProgressDlg->StopProgressDialog();
	}

}

BOOL
COM_CFileCopyingProgressDlg::IsCancelled()
{
	if ( NULL == m_ProgressDlg )
		return FALSE;

	return m_ProgressDlg->HasUserCancelled();
}

