/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef _PACKED_ARGS_H_
#define _PACKED_ARGS_H_

#include <windows.h>

#include <vector>
using namespace std;


#define PACKED_ARGS_SIG 0x1122

#include <pshpack1.h>

typedef struct PTR_STRUCT
{
    ULONG cbBuffer;
    PVOID pPtr;

	PTR_STRUCT() { pPtr = NULL; cbBuffer = 0; }
    
    PTR_STRUCT( ULONG _cbBuffer, PVOID _pPtr ) {pPtr =_pPtr;  cbBuffer = _cbBuffer; }


} PTR_STRUCT, *PPTR_STRUCT;


typedef struct PACKED_ARGS_HEADER
{
	WORD Signature;
	UINT Count;

} PACKED_ARGS_HEADER;

typedef struct PACKED_ARGS_ITEM
{
	CHAR   ArgType;
	ULONG  ArgSize;

} PACKED_ARGS_ITEM, *PPACKED_ARGS_ITEM;

#include <poppack.h>


class CPackedArgs 
{


typedef struct PTR_STRUCT2
{
    ULONG cbBuffer;
    PVOID pPtr;

} PTR_STRUCT2;



typedef struct ARG_STRUCT
{
	int ArgType;

	union
	{
		UINT    uArg;
		UINT64  uArg64;
		PTR_STRUCT2 PtrStruct;
	} ArgValue;

} ARG_STRUCT;


public:

    CPackedArgs();
    virtual ~CPackedArgs();


    static
    void *
    AllocateMem( ULONG MemSize );

	static 
	void
	FreeMem( PVOID pPtr );

	size_t
	GetCount();

	int
	GetArgType( UINT ArgNum );

	bool
	GetUInt32Arg( UINT ArgNum, UINT *pValue );

	bool
	GetUInt64Arg( UINT ArgNum, UINT64 *pValue );

	bool
	GetPtrArg( UINT ArgNum, PVOID *ppPtr, PULONG pcb );

	//fail if empty

	bool
	Pack( PVOID *pPtr, PULONG pSize );

	bool
	Unpack( PVOID pPtr, ULONG BuffSize );

	//PTR_STRUCT * for ARGT_PTR
	//UINT * for ARGT_UINT32
	//UINT64 * for ARGT_UINT64

	bool
	CreateFromArgs1(
		int ArgType1, 
		PVOID pArgValue1
		);

	bool
	CreateFromArgs2(
		int ArgType1, 
		PVOID pArgValue1,
		int ArgType2, 
		PVOID pArgValue2
		);

	bool
	CreateFromArgs3(
		int ArgType1, 
		PVOID pArgValue1,
		int ArgType2, 
		PVOID pArgValue2,
		int ArgType3, 
		PVOID pArgValue3
		);


public:

	static const int ARGT_UINT32         = 1;
	static const int ARGT_UINT64         = 2;
	static const int ARGT_PTR            = 3;
    static const int ARGT_DYNAMIC_PTR    = 4;
	static const ULONG MAX_ITEM_SIZE = 0x10000000;



private:

	bool
	PushArg(
		int ArgType, 
		PVOID pArgValue
		);


	void
	Empty();


private:


	vector<ARG_STRUCT> m_Args;

}; //CPackedArgs


#endif //_PACKED_ARGS_H_


