/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <windows.h>
#include "SecureFuncs.h"
#include <stdio.h>



#if defined(USE_INTERNAL_SECURE_FUNCS)



errno_t 
strcpy_s(
   char *strDestination,
   size_t numberOfElements,
   const char *strSource 
   )
{
    if ( strDestination == (void *)0 ||
         strSource == (void *)0
        )
    {
        return EINVAL;
    }

    if ( 0 == numberOfElements )
        return ERANGE;

    size_t i = (size_t)-1;

    do
    {
        i++;

        if ( (i >= numberOfElements - 1) && (strSource[ i ] != 0) )
        {
            strDestination[ i ] = 0;
            return ERANGE;
        }

        strDestination[ i ] = strSource[i];
        

    } while ( strSource[i] != 0 );

    return 0;
}

errno_t 
wcscpy_s(
   wchar_t *strDestination,
   size_t numberOfElements,
   const wchar_t *strSource 
    )
{
    if ( strDestination == (void *)0 ||
         strSource == (void *)0
        )
    {
        return EINVAL;
    }

    if ( 0 == numberOfElements )
        return ERANGE;

    size_t i = (size_t)-1;

    do
    {
        i++;

        if ( (i >= numberOfElements - 1) && (strSource[ i ] != 0) )
        {
            strDestination[ i ] = 0;
            return ERANGE;
        }

        strDestination[ i ] = strSource[i];
        

    } while ( strSource[i] != 0 );

    return 0;

}


errno_t
strncat_s(
   char *strDest,
   size_t numberOfElements,
   const char *strSource,
   size_t count
   )
{
    if ( strDest == (void *)0 ||
         strSource == (void *)0
        )
    {
        return EINVAL;
    }

    if ( 0 == numberOfElements )
        return ERANGE;

    size_t i = strlen(strDest);

    if ( (i + 1 >= count) && ( count != _TRUNCATE ) )
        return ERANGE;

    i -= 1;

    size_t k = (size_t)-1;


    do
    {
        i++;
        k++;

        if (  ( (i >= numberOfElements - 1) || 
                (i >= count - 1 && count != _TRUNCATE ) ) && 
            (strSource[ k ] != 0) 
            )
        {
            if ( i < numberOfElements )
                strDest[ i ] = 0;

            return ERANGE;
        }

        strDest[ i ] = strSource[k];


    } while ( strSource[k] != 0 );

    return 0;

}

errno_t
wcsncat_s(
   wchar_t *strDest,
   size_t numberOfElements,
   const wchar_t *strSource,
   size_t count 
  )
{
    if ( strDest == (void *)0 ||
         strSource == (void *)0
        )
    {
        return EINVAL;
    }

    if ( 0 == numberOfElements )
        return ERANGE;

    size_t i = wcslen(strDest);

    if ( (i + 1 >= count) && ( count != _TRUNCATE ) )
        return ERANGE;

    i -= 1;

    size_t k = (size_t)-1;

    do
    {
        i++;
        k++;

        if (  ( (i >= numberOfElements - 1) || 
                (i >= count - 1 && count != _TRUNCATE ) ) && 
            (strSource[ k ] != 0) 
            )
        {
            if ( i < numberOfElements )
                strDest[ i ] = 0;

            return ERANGE;
        }

        strDest[ i ] = strSource[k];


    } while ( strSource[k] != 0 );

    return 0;

}

int
_snprintf_s(
   char *buffer,
   size_t sizeOfBuffer,
   size_t count,
   const char *format,
   ...
   )
{
    size_t RealCount = 0;

    if ( count == _TRUNCATE )
        RealCount = sizeOfBuffer;
    else 
        RealCount = count < sizeOfBuffer ? count : sizeOfBuffer;

    if ( RealCount == 0 )
        return -1;

    va_list vaList;

    va_start( vaList, format );

    int ret = _vsnprintf(
        buffer,
        RealCount,
        format,
        vaList
        );

    if ( ret == -1 )
    {
        //TODO: set EINVAL
    }

    buffer[ RealCount - 1 ] = 0;
    return ret;
}


int 
_snwprintf_s(
   wchar_t *buffer,
   size_t sizeOfBuffer,
   size_t count,
   const wchar_t *format,
   ... 
   )
{
    size_t RealCount = 0;

    if ( count == _TRUNCATE )
        RealCount = sizeOfBuffer;
    else 
        RealCount = count < sizeOfBuffer ? count : sizeOfBuffer;

    if ( RealCount == 0 )
        return -1;

    va_list vaList;

    va_start( vaList, format );

    int ret = _vsnwprintf(
        buffer,
        RealCount,
        format,
        vaList
        );

    if ( ret == -1 )
    {
        //TODO: set EINVAL
    }

    buffer[ RealCount - 1 ] = 0;
    return ret;

}

int 
_vsnprintf_s(
   char *buffer,
   size_t sizeOfBuffer,
   size_t count,
   const char *format,
   va_list argptr 
  )
{
    size_t RealCount = 0;

    if ( count == _TRUNCATE )
        RealCount = sizeOfBuffer;
    else 
        RealCount = count < sizeOfBuffer ? count : sizeOfBuffer;

    if ( RealCount == 0 )
        return -1;

    int ret = _vsnprintf(
        buffer,
        RealCount,
        format,
        argptr
        );

    if ( ret == -1 )
    {
        //TODO: set EINVAL
    }

    buffer[ RealCount - 1 ] = 0;
    return ret;
}


int
_vsnwprintf_s(
   wchar_t *buffer,
   size_t sizeOfBuffer,
   size_t count,
   const wchar_t *format,
   va_list argptr 
    )
{
    size_t RealCount = 0;

    if ( count == _TRUNCATE )
        RealCount = sizeOfBuffer;
    else 
        RealCount = count < sizeOfBuffer ? count : sizeOfBuffer;

    if ( RealCount == 0 )
        return -1;

    int ret = _vsnwprintf(
        buffer,
        RealCount,
        format,
        argptr
        );

    if ( ret == -1 )
    {
        //TODO: set EINVAL
    }

    buffer[ RealCount - 1 ] = 0;
    return ret;

}

#endif
