/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "WinTools.h"
#include <shlobj.h>
#include <ShellApi.h>
#include "SecureFuncs.h"
#include "DbgPrintf.h"
#include "ChkMAlloc.h"

#include <strsafe.h> //StringCchCatA StringCchCopy

HANDLE g_AppMutex = NULL;

BOOL
EnumerateDirectoriesA(
	LPCSTR lpszDir,
	PVOID pContext,
	TEnumDirProcA pfnCallback
	)
{
	CHAR szMask[MAX_PATH+1] = "";
	size_t DirLen = strlen( lpszDir );
	size_t DirPos = 0;

	if ( DirLen >= ARRAYSIZE(szMask) )
		return FALSE;

	StringCchCopyA(
		szMask,
		ARRAYSIZE(szMask),
		lpszDir
		);

	if ( szMask[ DirLen - 1 ] != '\\' )
	{
		StringCchCatA(
			szMask,
			ARRAYSIZE(szMask),
			"\\*.*"
			);

		DirPos = DirLen + 1;
	}
	else
	{
		StringCchCatA(
			szMask,
			ARRAYSIZE(szMask),
			"*.*"
			);

		DirPos = DirLen;
	}

	WIN32_FIND_DATAA FindData;
	BOOL bRet = TRUE;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	for (
		hFind = FindFirstFileA(
			szMask,
			&FindData
			);
		bRet && hFind != INVALID_HANDLE_VALUE;
		bRet = FindNextFileA( hFind, &FindData )
		)
	{
		if ( 0 == _stricmp( FindData.cFileName, "." ) ||
			 0 == _stricmp( FindData.cFileName, ".." )
			 )
		{
			continue;
		}

		szMask[ DirPos ] = 0;

		StringCchCatA(
			szMask,
			ARRAYSIZE(szMask),
			FindData.cFileName
			);

		bRet = pfnCallback(
			pContext,
			szMask,
			FindData.dwFileAttributes,
			(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? 0 : 
			(ULONG64)FindData.nFileSizeLow | (((ULONG64)FindData.nFileSizeHigh) << 32)
			);

		if ( !bRet )
		{
			FindClose( hFind );
			return FALSE;
		}

		if ( FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			bRet = EnumerateDirectoriesA(
				szMask,
				pContext,
				pfnCallback
				);

			if ( !bRet )
			{
				FindClose( hFind );
				return FALSE;
			}
		}

	} //for
	

	FindClose( hFind );
	return TRUE;
}


BOOL
EnumerateDirectoriesW(
	LPCWSTR lpszDir,
	PVOID pContext,
	TEnumDirProcW pfnCallback
	)
{
	WCHAR szMask[MAX_PATH+1] = L"";
	size_t DirLen = wcslen( lpszDir );
	size_t DirPos = 0;

	if ( DirLen >= ARRAYSIZE(szMask) )
		return FALSE;

	StringCchCopyW(
		szMask,
		ARRAYSIZE(szMask),
		lpszDir
		);

	if ( szMask[ DirLen - 1 ] != L'\\' )
	{
		StringCchCatW(
			szMask,
			ARRAYSIZE(szMask),
			L"\\*.*"
			);

		DirPos = DirLen + 1;
	}
	else
	{
		StringCchCatW(
			szMask,
			ARRAYSIZE(szMask),
			L"*.*"
			);

		DirPos = DirLen;
	}

	WIN32_FIND_DATAW FindData;
	BOOL bRet = TRUE;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	for (
		hFind = FindFirstFileW(
			szMask,
			&FindData
			);
		bRet && hFind != INVALID_HANDLE_VALUE;
		bRet = FindNextFileW( hFind, &FindData )
		)
	{
		if ( 0 == _wcsicmp( FindData.cFileName, L"." ) ||
			 0 == _wcsicmp( FindData.cFileName, L".." )
			 )
		{
			continue;
		}

		szMask[ DirPos ] = 0;

		StringCchCatW(
			szMask,
			ARRAYSIZE(szMask),
			FindData.cFileName
			);

		bRet = pfnCallback(
			pContext,
			szMask,
			FindData.dwFileAttributes,
			(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? 0 : 
			(ULONG64)FindData.nFileSizeLow | (((ULONG64)FindData.nFileSizeHigh) << 32),
            &FindData.ftLastWriteTime
			);

		if ( !bRet )
		{
			FindClose( hFind );
			return FALSE;
		}

		if ( FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			bRet = EnumerateDirectoriesW(
				szMask,
				pContext,
				pfnCallback
				);

			if ( !bRet )
			{
				FindClose( hFind );
				return FALSE;
			}
		}

	} //for
	

	FindClose( hFind );
	return TRUE;

}


BOOLEAN
ClipboardGetCopiedFileList(
	vector<wstring> *pFileList
	)
{
	pFileList->clear();

	if ( !OpenClipboard( NULL ) )
		return FALSE;


	HGLOBAL hGlobal = GetClipboardData( CF_HDROP );

	if ( hGlobal == NULL )
	{
		CloseClipboard();
		return FALSE;
	}


	PVOID pBuffer = GlobalLock( hGlobal );

	if ( NULL == pBuffer )
	{
		CloseClipboard();
		return FALSE;
	}

	LPDROPFILES pDropFiles = (LPDROPFILES)pBuffer;
	LPCSTR pList = (LPCSTR)pBuffer + pDropFiles->pFiles;

	if ( pDropFiles->fWide )
	{
		LPWSTR lpwszList = (LPWSTR)pList;

		while ( lpwszList[0] != UNICODE_NULL )
		{
			pFileList->push_back( lpwszList );

			lpwszList += wcslen( lpwszList ) + 1;
		}
	}
	else
	{
		while ( pList[0] != 0 )
		{
			WCHAR wszFileName[ MAX_PATH ] = L"";

			int ret = MultiByteToWideChar(
				GetACP(),
				0,
				pList,
				-1,
				wszFileName,
				ARRAYSIZE( wszFileName )
				);

			if ( ret != 0 && wszFileName[0] != UNICODE_NULL )
			{
				pFileList->push_back( wszFileName );
			}

			pList += strlen( pList ) + 1;
		}

	}


	GlobalUnlock( hGlobal );
	CloseClipboard();

	return pFileList->size() != 0 ;
}

wstring
GetBaseFileNameW(
	LPCWSTR lpszFileName
	)
{
	wchar_t *psz = wcsrchr( (wchar_t *)lpszFileName, L'\\' );

	if ( psz )
		return psz+1;

	return lpszFileName;
}

wstring
GetFullPathFromRelative(
	LPCWSTR pwszRootDir,
	LPCWSTR pwszRelPath
    )
{
	wstring FullPath;

	FullPath = pwszRootDir;

	if ( FullPath.length() == 0 )
	{
		return L"";
	}

	if ( FullPath[ FullPath.length() - 1 ] != L'\\'  )
		FullPath += L"\\";

	if ( pwszRelPath[0] == L'\\' )
		FullPath += (pwszRelPath+1);
	else
		FullPath += (pwszRelPath);

    return FullPath;


}

BOOL
CreateFileWithRelPathW(
    HANDLE *pFileHandle,
    BOOL bOverWrite,
    wstring *pFileName,
	LPCWSTR pwszRootDir,
	LPCWSTR pwszRelPath,
	DWORD dwFileAttr
    )
{
	BOOL Result = FALSE;
	wstring FullPath;

	if ( pwszRootDir[0] == 0 )
	{
		return FALSE;
	}

    FullPath = GetFullPathFromRelative( pwszRootDir, pwszRelPath );

	if ( dwFileAttr & FILE_ATTRIBUTE_DIRECTORY )
	{
		Result = CreateDirectoryW(
			FullPath.c_str(),
			NULL
			);

		if ( !Result && GetLastError() == ERROR_PATH_NOT_FOUND )
		{
			if ( RecreatePathW( FullPath.c_str() ) )
			{
				Result = CreateDirectoryW(
					FullPath.c_str(),
					NULL
					);
			}
		}

		if ( !Result && ERROR_ALREADY_EXISTS == GetLastError() )
			Result = TRUE;
        
        if ( Result )
        {
            //restore attrs for folder?
            SetFileAttributesW( FullPath.c_str(), dwFileAttr ); 

            *pFileHandle = OpenDirW( FullPath.c_str(), FILE_WRITE_ATTRIBUTES );
        }

	}
	else
	{
		HANDLE hFile = CreateFileWithWriteAccessW(
			FullPath.c_str(),
			bOverWrite,
			dwFileAttr
			);

		if ( hFile == INVALID_HANDLE_VALUE &&
			GetLastError() == ERROR_PATH_NOT_FOUND
		   )
		{
			if ( RecreatePathW( FullPath.c_str() ) )
			{
				hFile = CreateFileWithWriteAccessW(
					FullPath.c_str(),
					bOverWrite,
					dwFileAttr
					);
			}
		}

		if ( hFile != INVALID_HANDLE_VALUE )
		{
			Result = TRUE;
			*pFileHandle = hFile;
		}
	}

	if ( Result )
	{
		 *pFileName = FullPath;
	}

	return Result;
}

BOOL
RecreatePathW(
    LPCWSTR lpszFileName
	)
{
	LPCWSTR psz = lpszFileName;

	if ( lpszFileName[0] == L'\\' )
	{
		psz++;

		if ( lpszFileName[1] == L'\\' )
			psz++;
	}

	while ( (psz = wcschr( psz, L'\\' )) != NULL )
	{
		wstring str;

		str.assign(
			lpszFileName,
			psz - lpszFileName
			);

		if ( (DWORD)-1 == GetFileAttributesW( str.c_str() ) )
		{
			BOOL bRet = CreateDirectoryW(
				str.c_str(),
				NULL
				);

			if ( !bRet )
				return FALSE;
		}

		psz++;

	}  //while

	return TRUE;
}

wstring
COM_LnkToFilePathW(
    LPCWSTR lpszLnkFile
    )
{
    IShellLinkW *ShellLinkObj = NULL;
    IPersistFile *PersistFileObj = NULL;
    wstring result;

    HRESULT hr = CoCreateInstance(
        CLSID_ShellLink,
        NULL, 
        CLSCTX_INPROC_SERVER,
        IID_IShellLinkW,
        (void**)&ShellLinkObj
        );     

    if ( !SUCCEEDED( hr ) )
        return L"";

    hr = ShellLinkObj->QueryInterface(
        IID_IPersistFile,
        (VOID **)&PersistFileObj
        );
        
    if ( !SUCCEEDED( hr ) )
    {
        ShellLinkObj->Release();
        return L"";
    }

    hr = PersistFileObj->Load(
        lpszLnkFile,
        STGM_READ
        );

    if ( SUCCEEDED( hr ) )
    {
        hr = ShellLinkObj->Resolve( 
            NULL,
            SLR_NO_UI | SLR_NOUPDATE | SLR_NOTRACK | SLR_NOLINKINFO | SLR_NOSEARCH
            );

        if ( SUCCEEDED( hr ) )
        {
            WCHAR wszPath[MAX_PATH] = L""; 

            hr = ShellLinkObj->GetPath(
                wszPath,
                ARRAYSIZE(wszPath),
                NULL, 
                0
                );        

            if ( SUCCEEDED( hr ) )
            {
                result = wszPath;
            }
        }
    }

    PersistFileObj->Release();
    ShellLinkObj->Release();
    return result;
}

BOOL
IsEqualFileExtW( LPCWSTR lpszFileName, LPCWSTR lpExt )
{
    int i = wcslen( lpszFileName ) - 1;
    
    while ( i >= 0 && lpszFileName[i] != L'\\' )
    {
        if ( lpszFileName[i] == L'.' )
        {
            if ( 0 == _wcsicmp( &lpszFileName[i], lpExt ) )
                return TRUE;
            else
                return FALSE;
        }

        i--;
    } // while

    return FALSE;

}

BOOLEAN
GetFileInfoW(
    LPCWSTR lpszFileName,
	UINT64 *FileSize,
	DWORD *pAttr,
    FILETIME *pFileLastWriteTime
	)
{
	WIN32_FIND_DATAW FindDataW;
	size_t len = wcslen( lpszFileName );


	if ( len == 0 )
    {
        DbgPrintf( "file name length is zero\n" );
		return FALSE;
    }

    RtlZeroMemory( &FindDataW, sizeof(FindDataW) );

	HANDLE hFind = INVALID_HANDLE_VALUE;

	if ( lpszFileName[len-1] == L'\\' )
	{
		//FindFirstFile fails if last char is '\\'

		WCHAR *pPath = (WCHAR *)MALLOC( (len+1)*sizeof(WCHAR) );

		if ( pPath == NULL )
        {
            DbgPrintf( "GetFileInfoW: No memory\n" );
			return FALSE;
        }

		StringCchCopyW( pPath, (len+1), lpszFileName );
		pPath[len-1] = UNICODE_NULL;

	    hFind = FindFirstFileW(
			pPath,
			&FindDataW
			);

		FREE( pPath );
        pPath = NULL;
	}
	else
	{		
	    hFind = FindFirstFileW(
			lpszFileName,
			&FindDataW
			);
	}

	if ( INVALID_HANDLE_VALUE == hFind )
    {
        DbgPrintf( "GetFileInfoW: FindFirstFileW failed\n" );
		return FALSE;
    }

	if ( FindDataW.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
	{
		*FileSize = 0;
	}
	else
	{
		*FileSize = ( (UINT64)FindDataW.nFileSizeLow | (((UINT64)FindDataW.nFileSizeHigh) << 32) );
	}

	*pAttr = FindDataW.dwFileAttributes;
    *pFileLastWriteTime = FindDataW.ftLastWriteTime;

	FindClose( hFind );
	return TRUE;

}

BOOL
OpenAppMutexAndCheckIsAppInstanceExist(
    LPCSTR lpAppMutexName
    )
{
     SECURITY_ATTRIBUTES SecAttr;
     UCHAR SecDescBuff[SECURITY_DESCRIPTOR_MIN_LENGTH];
     PSECURITY_DESCRIPTOR pSecDesc = (PSECURITY_DESCRIPTOR)SecDescBuff;

     if ( g_AppMutex )
        return FALSE;

     RtlZeroMemory( &SecAttr, sizeof(SecAttr) );

    InitializeSecurityDescriptor(
        pSecDesc,
        SECURITY_DESCRIPTOR_REVISION
        );

    SetSecurityDescriptorDacl(
        pSecDesc,
        TRUE,
        NULL,
        FALSE
        );

    SecAttr.nLength = sizeof(SecAttr);
    SecAttr.bInheritHandle = FALSE;
    SecAttr.lpSecurityDescriptor = pSecDesc;

    g_AppMutex = CreateMutex(
        &SecAttr,
        TRUE,
        lpAppMutexName
        );

    if ( g_AppMutex != NULL && ERROR_ALREADY_EXISTS == GetLastError() )
    {
        CloseHandle( g_AppMutex );
        g_AppMutex = NULL;
        return TRUE;
    }

    

    return FALSE;
}

BOOL
NotifyIcon_Add(
    HWND hWnd,
    ULONG uWindowMessage,
    HICON hIcon,
    LPCSTR lpszTip OPTIONAL
    )
{
    NOTIFYICONDATA NotifyIconData;

    RtlZeroMemory( &NotifyIconData, sizeof(NotifyIconData) );

    NotifyIconData.cbSize = sizeof( NotifyIconData );

    NotifyIconData.hWnd = hWnd;

    NotifyIconData.hIcon = hIcon;
    NotifyIconData.uCallbackMessage = uWindowMessage;


    NotifyIconData.uFlags = NIF_ICON | NIF_MESSAGE;

    if ( lpszTip )
    {
        NotifyIconData.uFlags |= NIF_TIP;

    	StringCchCopyA(
    		NotifyIconData.szTip,
    		ARRAYSIZE(NotifyIconData.szTip),
    		lpszTip
    		);
    }

    return Shell_NotifyIcon( NIM_ADD, &NotifyIconData );
}


BOOL
NotifyIcon_Delete(
    HWND hWnd
    )
{
    NOTIFYICONDATA NotifyIconData;

    RtlZeroMemory( &NotifyIconData, sizeof(NotifyIconData) );

    NotifyIconData.cbSize = sizeof( NotifyIconData );

    NotifyIconData.hWnd = hWnd;

    return Shell_NotifyIcon( NIM_DELETE, &NotifyIconData );

}

VOID
CloseAppMutex()
{
    if ( g_AppMutex )
    {
        CloseHandle( g_AppMutex );
        g_AppMutex = NULL;
    }
}

string
GetCurrentProcessDirA()
{
    CHAR szPath[ MAX_PATH + 1 ] = "";

    GetModuleFileName(
        GetModuleHandle( NULL ),
        szPath,
        ARRAYSIZE( szPath )
        );

    CHAR *psz = strrchr( szPath, '\\' );

    if ( psz )
        *psz = 0;

    return szPath;

}

string
GetCurrentProcessFileNameA()
{
    CHAR szPath[ MAX_PATH + 1 ] = "";

    GetModuleFileName(
        GetModuleHandle( NULL ),
        szPath,
        ARRAYSIZE( szPath )
        );

    CHAR *psz = strrchr( szPath, '\\' );

    if ( psz )
        return psz + 1;

    return szPath;
}

string
GetCurrentProcessFullPathA()
{
   CHAR szPath[ MAX_PATH + 1 ] = "";

    GetModuleFileName(
        GetModuleHandle( NULL ),
        szPath,
        ARRAYSIZE( szPath )
        );

    return szPath;
}

BOOL
IsFileExistA( LPCSTR lpFileName )
{
    DWORD dwAttr = GetFileAttributesA( lpFileName );

    if ( dwAttr == (DWORD)-1 )
        return FALSE;

    return TRUE;

}

BOOL
IsAdminGroupEnabled(
    BOOL *pEnabled
    )
{
    HANDLE hToken = NULL;

    if ( !OpenProcessToken( GetCurrentProcess(), TOKEN_QUERY, &hToken ) )
    {        
        return FALSE;
    }

    DWORD BuffSize = 0;

	GetTokenInformation( hToken, TokenGroups, 0, 0, &BuffSize );

    if ( 0 == BuffSize )
    {
        CloseHandle( hToken );
        return FALSE;
    }

    PVOID pBuffer = MALLOC( BuffSize );

    if ( NULL == pBuffer )
    {
        CloseHandle( hToken );
        return FALSE;
    }

    if ( !GetTokenInformation( hToken, TokenGroups, pBuffer, BuffSize, &BuffSize ) )
    {
        FREE( pBuffer );
        CloseHandle( hToken );
        return FALSE;
    }

    SID_IDENTIFIER_AUTHORITY NtAuth = { SECURITY_NT_AUTHORITY };
    PSID pAdminSID = NULL;

    if ( !AllocateAndInitializeSid(
            &NtAuth,
            2,
            SECURITY_BUILTIN_DOMAIN_RID,
            DOMAIN_ALIAS_RID_ADMINS,
            0,
            0,
            0,
            0,
            0,
            0,
            &pAdminSID )
       )
    {
        FREE( pBuffer );
        CloseHandle( hToken );
        return FALSE;
    }


    TOKEN_GROUPS *pTokenGroups = (TOKEN_GROUPS *)pBuffer;

    *pEnabled = FALSE;

    for ( ULONG u = 0; u < pTokenGroups->GroupCount; u++ )
    {
        if ( EqualSid( pAdminSID, pTokenGroups->Groups[ u ].Sid )  )
        {
            if ( pTokenGroups->Groups[ u ].Attributes & SE_GROUP_ENABLED )
            {
                *pEnabled = TRUE;
            }

            break;
        }
    }
    
    FreeSid( pAdminSID );
    FREE( pBuffer );
    CloseHandle( hToken );
    return TRUE;
}


BOOL
CALLBACK
FindFirstChild_EnumChildProc(
  HWND hwnd,
  LPARAM lParam
  )
{
	HWND *pWnd = (HWND *)lParam;

	*pWnd = hwnd;
	return FALSE;
}

HWND
FindFirstChildWindow( HWND hWnd )
{
	HWND hwndChild = NULL;

    EnumChildWindows(
		hWnd,
		FindFirstChild_EnumChildProc,
		(LPARAM)&hwndChild
		);

	return hwndChild;
}

wstring
GetDesktopFolder()
{
	WCHAR wszPath[ MAX_PATH ] = L"";

	LPITEMIDLIST pidlFolder = NULL;

	//Support old versions of windows

	HRESULT hr = SHGetSpecialFolderLocation(
		NULL,
		CSIDL_DESKTOPDIRECTORY,
		&pidlFolder		
		);

	if ( SUCCEEDED( hr ) )
	{
		SHGetPathFromIDListW( pidlFolder, wszPath );
		CoTaskMemFree( pidlFolder );
	}

	return wszPath;
}


string
UnicodeToAnsi(
    const wstring &WideStr
    )
{
    CHAR *pBuffer = (char *)MALLOC( WideStr.length() + 1 );
    
    if ( NULL == pBuffer )
        return "";

    *pBuffer = 0;
                
    WideCharToMultiByte(
        CP_ACP,
        0,
        WideStr.c_str(),
        WideStr.length(),
        pBuffer,
        WideStr.length() + 1,
        NULL,
        NULL
        );
        
    pBuffer[ WideStr.length() ] = 0;                
        
    string result = pBuffer; 
    FREE( pBuffer );
        
    return result;
}    

//return INVALID_HANDLE_VALUE if error
HANDLE
OpenDirW(
	LPCWSTR lpszDir,
	//FILE_LIST_DIRECTORY, FILE_WRITE_ATTRIBUTES
	DWORD dwAccess
	)
{
	HANDLE hDir = CreateFileW(
		lpszDir,
		dwAccess,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_BACKUP_SEMANTICS,
		NULL
		);

    return hDir;
}


string
FileTimeToString( FILETIME *pFileTime )
{
    SYSTEMTIME SysTime;

    FileTimeToSystemTime( pFileTime, &SysTime );

    CHAR szTime[ 100 ] = "";

    StringCchPrintfA(
        szTime,
        ARRAYSIZE(szTime),
        "%02u.%02u.%04u %u:%u:%u",
        (ULONG)SysTime.wDay,
        (ULONG)SysTime.wMonth,
        (ULONG)SysTime.wYear,
        (ULONG)SysTime.wHour,
        (ULONG)SysTime.wMinute,
        (ULONG)SysTime.wSecond
        );

    return string( szTime );
}

VOID
GetDirFilesListW(
    LPCWSTR lpszDir,
    vector<wstring> *pFilesList
    )
{
    pFilesList->clear();


	WCHAR szMask[MAX_PATH+1] = L"";
	size_t DirLen = wcslen( lpszDir );
	size_t DirPos = 0;

	if ( DirLen >= ARRAYSIZE(szMask) )
		return ;

	StringCchCopyW(
		szMask,
		ARRAYSIZE(szMask),
		lpszDir
		);

	if ( szMask[ DirLen - 1 ] != L'\\' )
	{
		StringCchCatW(
			szMask,
			ARRAYSIZE(szMask),
			L"\\*.*"
			);

		DirPos = DirLen + 1;
	}
	else
	{
		StringCchCatW(
			szMask,
			ARRAYSIZE(szMask),
			L"*.*"
			);

		DirPos = DirLen;
	}

	WIN32_FIND_DATAW FindData;
	BOOL bRet = TRUE;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	for (
		hFind = FindFirstFileW(
			szMask,
			&FindData
			);
		bRet && hFind != INVALID_HANDLE_VALUE;
		bRet = FindNextFileW( hFind, &FindData )
		)
	{
		if ( 0 == _wcsicmp( FindData.cFileName, L"." ) ||
			 0 == _wcsicmp( FindData.cFileName, L".." )
			 )
		{
			continue;
		}

		szMask[ DirPos ] = 0;

		StringCchCatW(
			szMask,
			ARRAYSIZE(szMask),
			FindData.cFileName
			);

        pFilesList->push_back( szMask );


	} //for
	

	FindClose( hFind );

}

BOOL
StringListToBuffW(
    vector<wstring> *pStrList,
    PVOID pBuff OPTIONAL,
    ULONG BuffSize,
    PULONG pRequiredBuffSize,
	PULONG pWritelnBytes
    )
{
    ULONG RequiredBuffSize = 0;
    
    RequiredBuffSize += sizeof(WCHAR)*2;

    for (
        vector<wstring>::iterator iter = pStrList->begin();
        iter != pStrList->end();
        iter++
        )
    {
        RequiredBuffSize += ( iter->length() + 1 ) * sizeof(WCHAR);
    }

	*pWritelnBytes = 0;
    *pRequiredBuffSize = RequiredBuffSize;

    if ( NULL == pBuff )
        return TRUE;

    if ( BuffSize < 2*sizeof(WCHAR) )
        return FALSE;

    PUCHAR pCurrentPtr = (PUCHAR)pBuff;
    ULONG CurrentBuffSize = BuffSize - sizeof(WCHAR)*2;

    for (
        vector<wstring>::iterator iter = pStrList->begin();
        iter != pStrList->end();
        iter++
        )
    {
        ULONG FullStringLengthInBytes = (iter->length() + 1)*sizeof(WCHAR);

        if ( FullStringLengthInBytes > CurrentBuffSize )
            break;

        RtlCopyMemory(
            pCurrentPtr,
            iter->c_str(),
            FullStringLengthInBytes
            );
		*pWritelnBytes += FullStringLengthInBytes;

        pCurrentPtr += FullStringLengthInBytes;
        CurrentBuffSize -= FullStringLengthInBytes;
    }


    RtlZeroMemory( pCurrentPtr, 2*sizeof(WCHAR) );
	*pWritelnBytes += 2*sizeof(WCHAR);

    return TRUE;
}

BOOL
SafeStringLenW(
    LPCWSTR lpszStr,
    ULONG StrWithNullCharMaxLen,
    PULONG pStrLen
    )
{
    if ( 0 == StrWithNullCharMaxLen )
        return FALSE;

    ULONG CurrentStrPos = 0;

    do
    {
        if ( lpszStr[CurrentStrPos] == 0 )
        {
            *pStrLen = CurrentStrPos;
            return TRUE;
        }

        CurrentStrPos++;
    } while ( CurrentStrPos < StrWithNullCharMaxLen );

    return FALSE;
}


BOOL
BuffToStringListW(
    vector<wstring> *pOutputStrList,
    PVOID pBuff,
    ULONG BuffSize
    )
{
    pOutputStrList->clear();

    if ( BuffSize < 2*sizeof(WCHAR) )
        return FALSE;


    PUCHAR pCurrentBuffPtr = (PUCHAR)pBuff;
    ULONG CurrentBuffSize = BuffSize - 2*sizeof(WCHAR);

    while ( CurrentBuffSize != 0 )
    {
        ULONG StrLen  = 0;

        BOOL bRet = SafeStringLenW(
            (LPCWSTR)pCurrentBuffPtr,
            (CurrentBuffSize / sizeof(WCHAR)),
            &StrLen
            );

        if ( !bRet )
        {
            pOutputStrList->clear();
            return FALSE;
        }

        pOutputStrList->push_back( (LPCWSTR)pCurrentBuffPtr );
        pCurrentBuffPtr += ( StrLen + 1 )*sizeof(WCHAR);
        CurrentBuffSize -= ( StrLen + 1 )*sizeof(WCHAR);
    }

    LPCWSTR lpszEnd = (LPCWSTR)pCurrentBuffPtr;

    if ( lpszEnd[0] != 0 || lpszEnd[1] != 0 )
        return FALSE;

    return TRUE;

}

HANDLE
CreateFileWithWriteAccessW(
	LPCWSTR lpszFileName,
	BOOL bOverWrite,
	DWORD dwFileAttr
	)
{

	HANDLE hFile = CreateFileW(
		lpszFileName,
		GENERIC_WRITE,
		FILE_SHARE_READ,
		NULL,
		bOverWrite ? CREATE_ALWAYS : CREATE_NEW,
		dwFileAttr,
		NULL
		);

	if ( hFile == INVALID_HANDLE_VALUE &&
		 bOverWrite
		 )
	{
		DWORD dwFileOldAttr = GetFileAttributesW( lpszFileName );

		if ( 
			dwFileOldAttr != (DWORD)-1 &&
			(dwFileOldAttr & ( FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM  )) )
		{
			if ( SetFileAttributesW( lpszFileName, dwFileOldAttr & ~(FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM) ) )
			{
				hFile = CreateFileW(
						lpszFileName,
						GENERIC_WRITE,
						FILE_SHARE_READ,
						NULL,
						bOverWrite ? CREATE_ALWAYS : CREATE_NEW,
						dwFileAttr,
						NULL
						);
			}
		}
	}

	return hFile;

}

//__success(ret==0)
int
RemoveDirectoryRecursiveW(
	IN LPCWSTR lpszDirFullPath,
	IN LPCWSTR lpszTitle OPTIONAL,
	IN BOOL bSilent
	)
{
	WCHAR szDirName[MAX_PATH+1];
	ZeroMemory( szDirName, sizeof(szDirName) );

	//(!)Showld be double null terminated
	StringCchCopyW( szDirName, ARRAYSIZE(szDirName) - 1, lpszDirFullPath );
	
	SHFILEOPSTRUCTW ShFileOp;

	ZeroMemory( &ShFileOp, sizeof(ShFileOp) );

	ShFileOp.hwnd = NULL;
	ShFileOp.wFunc = FO_DELETE;
	ShFileOp.pFrom = szDirName;
	ShFileOp.fFlags = FOF_NOCONFIRMATION | FOF_NOERRORUI;

	if ( bSilent )
		ShFileOp.fFlags |= FOF_SILENT;
		
	ShFileOp.lpszProgressTitle = lpszTitle;

	int ret = SHFileOperationW( &ShFileOp );

	return ret;
}

